"""Handler for wsi files using Pydicom."""

import json
from collections import defaultdict
from pathlib import Path
from typing import Any

import highdicom as hd
import numpy as np
import pydicom
import pydicom.errors
from pydicom.sr.codedict import codes
from pydicom.sr.coding import Code
from shapely.geometry import Polygon

from aignostics.utils import console, get_logger

logger = get_logger(__name__)


class PydicomHandler:
    def __init__(self, path: Path) -> None:
        self.path = Path(path)

    @classmethod
    def from_file(cls, path: str | Path) -> "PydicomHandler":
        """Create a PydicomHandler instance from a file or directory path.

        Args:
            path (str | Path): Path to the DICOM file or directory.

        Returns:
            PydicomHandler: An instance of PydicomHandler.
        """
        return cls(Path(path))

    def get_metadata(self, verbose: bool = False) -> dict[str, Any]:
        files = self._scan_files(verbose)
        return self._organize_by_hierarchy(files)

    def _scan_files(self, verbose: bool = False) -> list[dict[str, Any]]:  # noqa: C901, PLR0912, PLR0914, PLR0915
        dicom_files = []

        for file_path in self.path.rglob("*.dcm"):  # noqa: PLR1702
            if not file_path.is_file():
                continue

            try:
                print(file_path)
                ds = pydicom.dcmread(str(file_path), stop_before_pixels=True)
                # TODO(Helmut): Uncomment when DICOM is implemented
                # print(ds["Modality"].value)  # noqa: ERA001
                # print(getattr(ds, "Modality", "unknown"))  # noqa: ERA001
                # sys.exit()  # noqa: ERA001

                # Basic required DICOM fields
                file_info: dict[str, Any] = {
                    "path": str(file_path),
                    "study_uid": str(getattr(ds, "StudyInstanceUID", "unknown")),
                    "container_id": str(getattr(ds, "ContainerIdentifier", "unknown")),
                    "series_uid": str(getattr(ds, "SeriesInstanceUID", "unknown")),
                    "modality": str(getattr(ds, "Modality", "unknown")),
                    "type": "unknown",
                    "frame_of_reference_uid": str(getattr(ds, "FrameOfReferenceUID", "unknown")),
                }

                # Try to determine file type using highdicom
                try:
                    # TODO(Helmut): Check below, hd.sr.is_microscopy_bulk_simple_annotation(ds),
                    # hd.sr.is_microscopy_measurement(ds) for type annotation/measurement
                    if getattr(ds, "Modality", "") in {"SM", "WSI"}:
                        file_info["type"] = "image"
                except Exception:
                    logger.exception("Failed to analyze DICOM file with highdicom")
                    # If highdicom analysis fails, keep 'unknown' type

                # Add size and basic metadata
                file_info["size"] = file_path.stat().st_size
                file_info["metadata"] = {
                    "PatientID": str(getattr(ds, "PatientID", "unknown")),
                    "StudyDate": str(getattr(ds, "StudyDate", "unknown")),
                    "SeriesDescription": str(getattr(ds, "SeriesDescription", "")),
                }

                # Add to file_info dictionary after basic metadata
                if getattr(ds, "Modality", "") in {"SM", "WSI"}:
                    file_info.update({
                        "modality": getattr(ds, "Modality", ""),
                        "is_pyramidal": True,
                        "num_frames": int(getattr(ds, "NumberOfFrames", 1)),
                        "optical_paths": len(getattr(ds, "OpticalPathSequence", [])),
                        "focal_planes": len(getattr(ds, "DimensionIndexSequence", [])),
                        "total_pixel_matrix": (
                            int(getattr(ds, "TotalPixelMatrixColumns", 0)),
                            int(getattr(ds, "TotalPixelMatrixRows", 0)),
                        ),
                    })
                elif getattr(ds, "Modality", "") == "ANN":
                    ann = hd.ann.MicroscopyBulkSimpleAnnotations.from_dataset(ds)
                    group_infos = []
                    groups = ann.get_annotation_groups()
                    for group in groups:
                        # Calculate min/max coordinates for all points
                        col_min = row_min = float("inf")  # Initialize to positive infinity
                        col_max = row_max = float("-inf")  # Initialize to negative infinity
                        graphic_data_len = float("-inf")
                        first = None

                        if verbose:
                            graphic_data = group.get_graphic_data(ann.annotation_coordinate_type)
                            graphic_data_len = len(graphic_data)
                            first = graphic_data[0]
                            if graphic_data:
                                if group.graphic_type == hd.ann.GraphicTypeValues.POINT:
                                    # For points, each item is a single coordinate
                                    for point in graphic_data:
                                        col_min = min(col_min, point[0])
                                        col_max = max(col_max, point[0])
                                        row_min = min(row_min, point[1])
                                        row_max = max(row_max, point[1])
                                else:
                                    # For polygons/polylines, process all polygons
                                    for polygon in graphic_data:
                                        for point in polygon:
                                            col_min = min(col_min, point[0])
                                            col_max = max(col_max, point[0])
                                            row_min = min(row_min, point[1])
                                            row_max = max(row_max, point[1])

                        group_infos.append({
                            "uid": group.uid,
                            "label": group.label,
                            "property_type": group.annotated_property_type,
                            "graphic_type": group.graphic_type,
                            "count": graphic_data_len,
                            "first": first,
                            "col_min": float(col_min),
                            "row_min": float(row_min),
                            "col_max": float(col_max),
                            "row_max": float(row_max),
                        })
                    file_info.update({
                        "modality": "ANN",
                        "coordinate_type": ann.annotation_coordinate_type,
                        "annotation_groups": group_infos,
                    })

                # Extract pyramid levels from frame organization
                if getattr(ds, "DimensionOrganizationSequence", None):
                    # Get frame organization
                    frame_groups = {}
                    for frame in getattr(ds, "PerFrameFunctionalGroupsSequence", []):
                        level_idx = frame.DimensionIndexValues[0]
                        if level_idx not in frame_groups:
                            frame_groups[level_idx] = {
                                "count": 0,
                                "rows": int(frame.get("Rows", 0)),
                                "columns": int(frame.get("Columns", 0)),
                            }
                        frame_groups[level_idx]["count"] += 1

                    # Sort and store pyramid level information
                    pyramid_info = [
                        {
                            "level": int(level_idx),
                            "frame_count": frame_groups[level_idx]["count"],
                            "frame_size": (
                                frame_groups[level_idx]["columns"],
                                frame_groups[level_idx]["rows"],
                            ),
                        }
                        for level_idx in sorted(frame_groups.keys())
                    ]
                    file_info["pyramid_info"] = pyramid_info

                dicom_files.append(file_info)

            except pydicom.errors.InvalidDicomError:
                continue

        return dicom_files

    @staticmethod
    def _organize_by_hierarchy(files: list[dict[str, Any]]) -> dict[str, Any]:
        if not files:
            return {"type": "empty", "message": "No DICOM files found"}

        if len(files) == 1:
            return {"type": "file", "file_info": files[0]}

        # Group by study -> container -> series
        studies = defaultdict(
            lambda: {
                "study_info": {
                    "study_uid": "",
                    "study_id": "",
                    "study_date": "",
                    "study_time": "",
                    "accession_number": "",
                },
                "patient_info": {"id": "", "name": "", "gender": "", "birth_date": ""},
                "clinical_trial": {
                    "sponsor_name": "",
                    "protocol_id": "",
                    "protocol_name": "",
                    "site_name": "",
                },
                "slides": defaultdict(
                    lambda: {
                        "specimen_info": {
                            "description": "",
                            "anatomical_structure": "",
                            "collection_method": "",
                            "parent_specimens": [],
                            "embedding_medium": "",
                        },
                        "equipment_info": {
                            "manufacturer": "",
                            "model_name": "",
                            "device_serial_number": "",
                            "software_version": "",
                            "institution_name": "",
                        },
                        "series": defaultdict(lambda: {"description": "", "modality": "", "files": []}),
                    }
                ),
            }
        )

        for file_info in files:
            study_uid = file_info["study_uid"]
            container_id = file_info["container_id"]
            series_uid = file_info["series_uid"]
            ds = pydicom.dcmread(file_info["path"], stop_before_pixels=True)

            # Update study info if not already set
            if not studies[study_uid]["study_info"]["study_id"]:
                studies[study_uid]["study_info"].update({
                    "study_uid": study_uid,
                    "study_id": str(getattr(ds, "StudyID", "")),
                    "study_date": str(getattr(ds, "StudyDate", "")),
                    "study_time": str(getattr(ds, "StudyTime", "")),
                    "accession_number": str(getattr(ds, "AccessionNumber", "")),
                })

                # Update patient info
                studies[study_uid]["patient_info"].update({
                    "id": str(getattr(ds, "PatientID", "")),
                    "name": str(getattr(ds, "PatientName", "")),
                    "gender": str(getattr(ds, "PatientSex", "")),
                    "birth_date": str(getattr(ds, "PatientBirthDate", "")),
                })

                # Update clinical trial info
                studies[study_uid]["clinical_trial"].update({
                    "sponsor_name": str(getattr(ds, "ClinicalTrialSponsorName", "")),
                    "protocol_id": str(getattr(ds, "ClinicalTrialProtocolID", "")),
                    "protocol_name": str(getattr(ds, "ClinicalTrialProtocolName", "")),
                    "site_name": str(getattr(ds, "ClinicalTrialSiteName", "")),
                })

            # Update series info if not already set
            series = studies[study_uid]["slides"][container_id]["series"][series_uid]
            if not series["description"]:
                series.update({
                    "description": str(getattr(ds, "SeriesDescription", "")),
                    "modality": str(getattr(ds, "Modality", "")),
                })

            # Add file-specific info only
            file_specific = {
                "path": file_info["path"],
                "size": Path(file_info["path"]).stat().st_size,
                "instance_uid": str(getattr(ds, "SOPInstanceUID", "")),
                "frame_of_reference_uid": str(getattr(ds, "FrameOfReferenceUID", "")),
                "type": file_info["type"],
                "dimensions": None,  # Initialize dimensions as None,
            }

            # Add generic image dimensions for any image type
            if hasattr(ds, "Rows") and hasattr(ds, "Columns"):
                file_specific["dimensions"] = (int(ds.Rows), int(ds.Columns))
                file_specific["photometric_interpretation"] = str(getattr(ds, "PhotometricInterpretation", ""))
                file_specific["bits_allocated"] = int(getattr(ds, "BitsAllocated", 0))
                file_specific["bits_stored"] = int(getattr(ds, "BitsStored", 0))
                file_specific["samples_per_pixel"] = int(getattr(ds, "SamplesPerPixel", 0))
                file_specific["image_type"] = getattr(ds, "ImageType", [])

            # Copy pyramidal information if present
            if file_info.get("is_pyramidal"):
                file_specific.update({
                    "is_pyramidal": True,
                    "num_frames": file_info["num_frames"],
                    "optical_paths": file_info["optical_paths"],
                    "focal_planes": file_info["focal_planes"],
                    "total_pixel_matrix": file_info["total_pixel_matrix"],
                })
                if file_info.get("pyramid_info"):
                    file_specific["pyramid_info"] = file_info["pyramid_info"]
            else:
                file_specific.update(file_info)

            series["files"].append(file_specific)

            # Update the specimen and equipment info when processing files
            if not studies[study_uid]["slides"][container_id]["specimen_info"]["description"]:
                studies[study_uid]["slides"][container_id]["specimen_info"].update({
                    "description": str(
                        getattr(ds, "SpecimenDescriptionSequence", [""])[0].get("SpecimenShortDescription", "")
                        if getattr(ds, "SpecimenDescriptionSequence", [])
                        else ""
                    ),
                    "anatomical_structure": str(
                        getattr(ds, "SpecimenDescriptionSequence", [""])[0]
                        .get("PrimaryAnatomicStructureSequence", [{}])[0]
                        .get("CodeMeaning", "")
                        if getattr(ds, "SpecimenDescriptionSequence", [])
                        else ""
                    ),
                    "collection_method": str(
                        getattr(ds, "SpecimenDescriptionSequence", [""])[0].get(
                            "SpecimenCollectionProcedureDescription", ""
                        )
                        if getattr(ds, "SpecimenDescriptionSequence", [])
                        else ""
                    ),
                    "parent_specimens": [
                        str(x.get("SpecimenIdentifier", ""))
                        for x in getattr(ds, "SpecimenDescriptionSequence", [])
                        if x.get("SpecimenIdentifier")
                    ],
                    "embedding_medium": str(
                        getattr(ds, "SpecimenDescriptionSequence", [""])[0].get("SpecimenEmbeddingMethod", "")
                        if getattr(ds, "SpecimenDescriptionSequence", [])
                        else ""
                    ),
                })

                studies[study_uid]["slides"][container_id]["equipment_info"].update({
                    "manufacturer": str(getattr(ds, "Manufacturer", "")),
                    "model_name": str(getattr(ds, "ManufacturerModelName", "")),
                    "device_serial_number": str(getattr(ds, "DeviceSerialNumber", "")),
                    "software_version": str(getattr(ds, "SoftwareVersions", "")),
                    "institution_name": str(getattr(ds, "InstitutionName", "")),
                })

        return {"type": "root", "studies": studies}

    def __enter__(self) -> "PydicomHandler":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:  # noqa: ANN001
        return

    @staticmethod
    def geojson_import(dicom_path: Path, geojson_path: Path) -> bool:  # noqa: C901, PLR0912, PLR0914, PLR0915
        """Convert GeoJSON to DICOM ANN instance.

        Args:
            dicom_path (Path): Path to the DICOM directory.
            geojson_path (Path): Path to the GeoJSON file.

        Returns:
            bool: True if the import was successful, False otherwise.
        """
        try:
            with open(geojson_path, encoding="utf-8") as f:
                geojson_data = json.load(f)

            largest_file = None
            largest_dimension = 0

            # determine the largest image in the directory
            for file_path in dicom_path.rglob("*"):
                if not file_path.is_file():
                    continue

                try:
                    ds = pydicom.dcmread(str(file_path), stop_before_pixels=True)
                    if getattr(ds, "Modality", "") in {"SM", "WSI"}:
                        columns = int(getattr(ds, "TotalPixelMatrixColumns", 0))
                        rows = int(getattr(ds, "TotalPixelMatrixRows", 0))
                        dimension = columns * rows

                        if dimension > largest_dimension:
                            largest_dimension = dimension
                            largest_file = file_path

                except pydicom.errors.InvalidDicomError as e:
                    console.print(f"Failed to process file {file_path}: {e}")
                    continue

            if largest_file:
                ds = pydicom.dcmread(str(largest_file), stop_before_pixels=True)
                columns = int(getattr(ds, "TotalPixelMatrixColumns", 0))
                rows = int(getattr(ds, "TotalPixelMatrixRows", 0))
                graphic_data = []
                graphic_types = []
                area_measurement_values = []

                for feature in geojson_data["features"]:
                    # We consider the outer geometry only,
                    # not additional in properties of a feature (as is used for the "cell" objectType)
                    geometry = feature["geometry"]

                    if geometry["type"] == "Point":
                        coordinates = np.array(geometry["coordinates"], dtype=np.float32)
                        if not (0 <= coordinates[0] < columns and 0 <= coordinates[1] < rows):
                            console.print(f"Point coordinates {coordinates} out of bounds")
                            continue
                        graphic_data.append(coordinates)
                        graphic_types.append(hd.ann.GraphicTypeValues.POINT)

                    elif geometry["type"] == "Polygon":
                        # DICOM does only contain simple polygons, without holes
                        coordinates = np.array(geometry["coordinates"][0], dtype=np.float32)
                        # Remove last point if it's identical to first (closed polygon)
                        if np.array_equal(coordinates[0], coordinates[-1]):
                            coordinates = coordinates[:-1]

                        # convert to use shapely
                        polygon = Polygon(coordinates)

                        # Check if enough points remain for valid polygon
                        if not polygon.is_valid:
                            console.print("Not a valid polygon")
                            continue

                        # Check if coordinates are within bounds
                        in_bounds = Polygon([
                            (0, 0),
                            (columns, 0),
                            (columns, rows),
                            (0, rows),
                        ]).contains(polygon)
                        if not in_bounds:
                            continue

                        # Add polygon data
                        graphic_data.append(coordinates)
                        graphic_types.append(hd.ann.GraphicTypeValues.POLYGON)

                        # Add measurements
                        area_measurement_values.append(np.float32(polygon.area))

                    else:
                        continue

                area_measurement = hd.ann.Measurements(
                    name=codes.SCT.Area,
                    unit=codes.UCUM.SquareMicrometer,
                    values=np.array(area_measurement_values, dtype=np.float32),
                )

                annotation_group = hd.ann.AnnotationGroup(
                    number=1,
                    uid=pydicom.uid.generate_uid(),  # type: ignore
                    label="Cell nuclei",
                    description="Generated by Orion CLI",
                    annotated_property_category=codes.SCT.AnatomicalStructure,
                    annotated_property_type=Code(
                        "84640000", "SCT", "Nucleus"
                    ),  # https://termbrowser.nhs.uk/?perspective=full&conceptId1=84640000&edition=uk-edition&release=v20240925&server=https://termbrowser.nhs.uk/sct-browser-api/snomed&langRefset=999001261000000100,999000691000001104
                    algorithm_type=hd.ann.AnnotationGroupGenerationTypeValues.AUTOMATIC,
                    algorithm_identification=hd.AlgorithmIdentificationSequence(
                        "aignx:heta",
                        hd.sr.CodedConcept(  # https://dicom.nema.org/medical/dicom/current/output/chtml/part16/sect_cid_7162.html
                            "123110", "DCM", "Artificial Intelligence"
                        ),
                        "0.0.1",
                        source="Aignostics GmbH (https://www.aignostics.com)",
                        parameters={"runtime": "PAPI"},
                    ),
                    graphic_type=(graphic_types[0] if graphic_types else hd.ann.GraphicTypeValues.POINT),
                    graphic_data=graphic_data,
                    measurements=[area_measurement],
                )

                bulk_annotations = hd.ann.MicroscopyBulkSimpleAnnotations(
                    source_images=[ds],
                    annotation_coordinate_type=hd.ann.AnnotationCoordinateTypeValues.SCOORD,
                    annotation_groups=[annotation_group],
                    series_instance_uid=pydicom.uid.generate_uid(),
                    series_number=10,
                    sop_instance_uid=pydicom.uid.generate_uid(),
                    instance_number=1,
                    manufacturer="Aignostics GmbH",
                    manufacturer_model_name="aignx:heta",
                    software_versions="0.0.1",
                    device_serial_number="1234",
                    content_description=f"{geojson_path.stem} Annotations",
                )

                output_filename = f"{geojson_path.stem}_annotations.dcm"
                bulk_annotations.save_as(str(dicom_path / output_filename))

            return True

        except (OSError, json.JSONDecodeError) as e:
            console.print(f"Failed to import GeoJSON: {e}")
            return False



"""Boot sequence."""

import os
import sys
from pathlib import Path

from ._log import logging_initialize

_boot_called = False

# Import third party dependencies
third_party_dir = Path(__file__).parent.absolute() / ".." / "third_party"
if third_party_dir.is_dir() and str(third_party_dir) not in sys.path:
    sys.path.insert(0, str(third_party_dir))


def boot(modules_to_instrument: list[str]) -> None:
    """Boot the application.

    Args:
        modules_to_instrument (list): List of modules to be instrumented.
        repository_url (str): URL of the repository.
        repository_root_path (str): The root path of the repository. Default is the root path.
    """
    global _boot_called  # noqa: PLW0603
    if _boot_called:
        return
    _boot_called = True

    from ._sentry import sentry_initialize  # noqa: PLC0415

    sentry_initialize()

    log_to_logfire = False
    from ._logfire import logfire_initialize  # noqa: PLC0415

    log_to_logfire = logfire_initialize(modules_to_instrument)

    _parse_env_args()
    logging_initialize(log_to_logfire)
    _amend_library_path()
    _log_boot_message()


from ._constants import __project_name__, __version__  # noqa: E402
from ._log import get_logger  # noqa: E402
from ._process import get_process_info  # noqa: E402


def _parse_env_args() -> None:
    """Parse --env arguments from command line and add to environment if prefix matches.

    - Last but not least removes those args so typer does not complain about them.
    """
    i = 1  # Start after script name
    to_remove = []
    prefix = f"{__project_name__.upper()}_"

    while i < len(sys.argv):
        current_arg = sys.argv[i]

        # Handle "--env KEY=VALUE" or "-e KEY=VALUE" format (two separate arguments)
        if (current_arg in {"--env", "-e"}) and i + 1 < len(sys.argv):
            key_value = sys.argv[i + 1]
            if "=" in key_value:
                key, value = key_value.split("=", 1)
                if key.startswith(prefix):
                    os.environ[key] = value.strip("\"'")
                to_remove.extend([i, i + 1])
                i += 2
                continue

        i += 1

    # Remove processed arguments from sys.argv in reverse order
    for index in sorted(to_remove, reverse=True):
        del sys.argv[index]


def _amend_library_path() -> None:
    """Patch environment variables before any other imports."""
    if "DYLD_FALLBACK_LIBRARY_PATH" not in os.environ:
        os.environ["DYLD_FALLBACK_LIBRARY_PATH"] = f"{os.getenv('HOMEBREW_PREFIX', '/opt/homebrew')}/lib/"


def _log_boot_message() -> None:
    """Log boot message with version and process information."""
    logger = get_logger(__name__)
    process_info = get_process_info()
    logger.info(
        "⭐ Booting %s v%s (project root %s, pid %s), parent '%s' (pid %s)",
        __project_name__,
        __version__,
        process_info.project_root,
        process_info.pid,
        process_info.parent.name,
        process_info.parent.pid,
    )



"""
This module provides the main client interface for interacting with Aignostics services.

It offers functionality for authentication, data management, and API operations.
The primary class in this module is the `Client` class, which serves as the entry point
for all interactions with the Aignostics platform.
"""

from aignx.codegen.exceptions import ApiException, NotFoundException
from aignx.codegen.models import ApplicationReadResponse as Application
from aignx.codegen.models import ApplicationRunStatus, ItemStatus
from aignx.codegen.models import ApplicationVersionReadResponse as ApplicationVersion
from aignx.codegen.models import InputArtifactCreationRequest as InputArtifact
from aignx.codegen.models import InputArtifactReadResponse as InputArtifactData
from aignx.codegen.models import ItemCreationRequest as InputItem
from aignx.codegen.models import ItemResultReadResponse as ItemResult
from aignx.codegen.models import OutputArtifactReadResponse as OutputArtifactData
from aignx.codegen.models import OutputArtifactResultReadResponse as OutputArtifactElement
from aignx.codegen.models import RunReadResponse as ApplicationRunData

from ._cli import cli
from ._client import Client
from ._constants import (
    API_ROOT_DEV,
    API_ROOT_PRODUCTION,
    API_ROOT_STAGING,
    AUDIENCE_DEV,
    AUDIENCE_PRODUCTION,
    AUDIENCE_STAGING,
    AUTHORIZATION_BASE_URL_DEV,
    AUTHORIZATION_BASE_URL_PRODUCTION,
    AUTHORIZATION_BASE_URL_STAGING,
    DEVICE_URL_DEV,
    DEVICE_URL_PRODUCTION,
    DEVICE_URL_STAGING,
    JWS_JSON_URL_DEV,
    JWS_JSON_URL_PRODUCTION,
    JWS_JSON_URL_STAGING,
    REDIRECT_URI_DEV,
    REDIRECT_URI_PRODUCTION,
    REDIRECT_URI_STAGING,
    TOKEN_URL_DEV,
    TOKEN_URL_PRODUCTION,
    TOKEN_URL_STAGING,
)
from ._messages import AUTHENTICATION_FAILED, NOT_YET_IMPLEMENTED, UNKNOWN_ENDPOINT_URL
from ._service import Service, TokenInfo, UserInfo
from ._settings import Settings, settings
from ._utils import calculate_file_crc32c, download_file, generate_signed_url, mime_type_to_file_ending
from .resources.runs import LIST_APPLICATION_RUNS_MAX_PAGE_SIZE, LIST_APPLICATION_RUNS_MIN_PAGE_SIZE, ApplicationRun

__all__ = [
    "API_ROOT_DEV",
    "API_ROOT_PRODUCTION",
    "API_ROOT_STAGING",
    "AUDIENCE_DEV",
    "AUDIENCE_PRODUCTION",
    "AUDIENCE_STAGING",
    "AUTHENTICATION_FAILED",
    "AUTHORIZATION_BASE_URL_DEV",
    "AUTHORIZATION_BASE_URL_PRODUCTION",
    "AUTHORIZATION_BASE_URL_STAGING",
    "DEVICE_URL_DEV",
    "DEVICE_URL_PRODUCTION",
    "DEVICE_URL_STAGING",
    "JWS_JSON_URL_DEV",
    "JWS_JSON_URL_PRODUCTION",
    "JWS_JSON_URL_STAGING",
    "LIST_APPLICATION_RUNS_MAX_PAGE_SIZE",
    "LIST_APPLICATION_RUNS_MIN_PAGE_SIZE",
    "NOT_YET_IMPLEMENTED",
    "NOT_YET_IMPLEMENTED",
    "REDIRECT_URI_DEV",
    "REDIRECT_URI_PRODUCTION",
    "REDIRECT_URI_STAGING",
    "TOKEN_URL_DEV",
    "TOKEN_URL_PRODUCTION",
    "TOKEN_URL_STAGING",
    "UNKNOWN_ENDPOINT_URL",
    "ApiException",
    "Application",
    "ApplicationRun",
    "ApplicationRunData",
    "ApplicationRunStatus",
    "ApplicationRunStatus",
    "ApplicationVersion",
    "Client",
    "InputArtifact",
    "InputArtifactData",
    "InputItem",
    "ItemResult",
    "ItemStatus",
    "NotFoundException",
    "OutputArtifactData",
    "OutputArtifactElement",
    "Service",
    "Settings",
    "TokenInfo",
    "UserInfo",
    "calculate_file_crc32c",
    "cli",
    "download_file",
    "generate_signed_url",
    "mime_type_to_file_ending",
    "settings",
]



"""Messages for the Aignostics client."""

AUTHENTICATION_FAILED = "Authentication failed. Please check your credentials."
NOT_YET_IMPLEMENTED = "Not yet implemented."
UNKNOWN_ENDPOINT_URL = "Unknown endpoint URL. Please check the API documentation."
INVALID_REDIRECT_URI = "Invalid redirect URI. Please check the redirect URI in your application settings."



"""Constants generated from the environment at runtime used throughout the project."""

import os
import sys
from importlib import metadata
from pathlib import Path

from dotenv import load_dotenv

__project_name__ = __name__.split(".")[0]
load_dotenv(str(Path(".env")))
load_dotenv(os.getenv(f"{__project_name__.upper()}_ENV_FILE", Path.home() / f".{__project_name__}/.env"))

__project_path__ = str(Path(__file__).parent.parent.parent)
__version__ = metadata.version(__project_name__)
__is_development_mode__ = "uvx" not in sys.argv[0].lower()
__is_running_in_container__ = os.getenv(f"{__project_name__.upper()}_RUNNING_IN_CONTAINER")

# Determine if we're running in a read-only runtime environment
READ_ONLY_ENV_INDICATORS = [
    f"{__project_name__.upper()}_RUNNING_IN_CONTAINER",
    "VERCEL_ENV",
    "RAILWAY_ENVIRONMENT",
]
__is_running_in_read_only_environment__ = any(os.getenv(env_var) is not None for env_var in READ_ONLY_ENV_INDICATORS)

# Determine environment we are deployed on
ENV_VAR_MAPPINGS = {
    "ENV": lambda env: env,
    "VERCEL_ENV": lambda env: env,  # See https://vercel.com/docs/environment-variables/system-environment-variables
    "RAILWAY_ENVIRONMENT": lambda env: env,  # See https://docs.railway.com/reference/variables#railway-provided-variables
}
__env__ = "local"  # Default
for env_var, mapper in ENV_VAR_MAPPINGS.items():
    env_value = os.getenv(env_var)
    if env_value:
        __env__ = mapper(env_value)  # type: ignore[no-untyped-call]
        break

# Define environment file paths
__env_file__ = [
    Path.home() / f".{__project_name__}" / ".env",
    Path.home() / f".{__project_name__}" / f".env.{__env__}",
    Path(".env"),
    Path(f".env.{__env__}"),
]
env_file_path = os.getenv(f"{__project_name__.upper()}_ENV_FILE")
if env_file_path:
    __env_file__.insert(2, Path(env_file_path))

# Determine __base_url__
PLATFORM_URL_MAPPINGS = {
    "VERCEL_URL": lambda url: f"https://{url}",  # See https://vercel.com/docs/environment-variables/system-environment-variables
    "RAILWAY_PUBLIC_DOMAIN": lambda url: f"https://{url}",  # See https://docs.railway.com/reference/variables#railway-provided-variables
}
__base__url__ = os.getenv(f"{__project_name__.upper()}_BASE_URL")
if not __base__url__:
    for env_var, mappers in PLATFORM_URL_MAPPINGS.items():
        env_value = os.getenv(env_var)
        if env_value:
            __base__url__ = mappers(env_value)  # type: ignore[no-untyped-call]
            break


def get_project_url_by_label(prefix: str) -> str:
    """Get labeled Project-URL.

    See https://packaging.python.org/en/latest/specifications/core-metadata/#core-metadata-project-url

    Args:
        prefix(str): The prefix to match at the beginning of URL entries.

    Returns:
        The extracted URL string if found, or an empty string if not found.
    """
    for url_entry in metadata.metadata(__project_name__).get_all("Project-URL", []):
        if url_entry.startswith(prefix):
            return str(url_entry.split(", ", 1)[1])
    return ""


_authors = metadata.metadata(__project_name__).get_all("Author-email", [])
_author = _authors[0] if _authors else None
__author_name__ = _author.split("<")[0].strip() if _author else None
__author_email__ = _author.split("<")[1].strip(" >") if _author else None
__repository_url__ = get_project_url_by_label("Source")
__documentation__url__ = get_project_url_by_label("Documentation")



"""Layout including sidebar and menu."""

import contextlib
import platform
from collections.abc import Generator
from contextlib import contextmanager
from importlib.util import find_spec
from typing import Any

from humanize import naturaldelta
from nicegui import app, ui

from aignostics.utils import __version__, open_user_data_directory

from ._theme import theme

FLAT_COLOR_WHITE = "flat color=white"

HEALTH_UPDATE_INTERVAL = 30
USERINFO_UPDATE_INTERVAL = 60


@contextmanager
def frame(  # noqa: C901, PLR0915
    navigation_title: str,
    navigation_icon: str | None = None,
    navigation_icon_color: str | None = None,
    navigation_icon_tooltip: str | None = None,
    left_sidebar: bool = False,
) -> Generator[Any, Any, Any]:
    """Custom page frame to share the same styling and behavior across all pages.

    Args:
        navigation_title (str): The title of the navigation bar.
        navigation_icon (str | None): The icon for the navigation bar.
        navigation_icon_color (str | None): The color of the navigation icon.
        navigation_icon_tooltip (str | None): The tooltip for the navigation icon.
        left_sidebar (bool): Whether to show the left sidebar or not.

    Yields:
        Generator[Any, Any, Any]: The context manager for the page frame.
    """
    from nicegui import background_tasks, context, run  # noqa: PLC0415

    from aignostics.platform import Service as PlatformService  # noqa: PLC0415
    from aignostics.platform import UserInfo  # noqa: PLC0415
    from aignostics.system import Service as SystemService  # noqa: PLC0415

    theme()

    user_info: UserInfo | None = None
    launchpad_healthy: bool | None = None

    @ui.refreshable
    def _user_info_ui() -> None:
        spinner = ui.spinner().props("flat color=purple-400")
        if user_info and user_info.profile:
            spinner.set_visibility(False)
            icon = "img:" + user_info.profile.picture if user_info.profile.picture else "account_circle"
            with (
                ui.dropdown_button(icon=icon)
                .style("width: 30px; height: 30px; border-radius: 50%")
                .props(FLAT_COLOR_WHITE),
                ui.card(),
            ):
                ui.label(f"{user_info.profile.name} ({user_info.profile.email})")
                ui.label(f"Organisation: {user_info.org_name or user_info.org_id}")
                ui.label(f"Role: {user_info.role}")
                ui.label(f"Token Expiry: {naturaldelta(user_info.token.expires_in)}")
                ui.separator()
                with ui.row().classes("items-center justify-between"):
                    ui.button("Re-authenticate", icon="switch_account", on_click=_user_info_ui_relogin).props(
                        add="style=fab-mini"
                    )

    async def _user_info_ui_load() -> None:
        nonlocal user_info
        with contextlib.suppress(Exception):
            user_info = await run.cpu_bound(PlatformService.get_user_info)
        await ui.context.client.connected()
        app.storage.tab["user_info"] = user_info
        _user_info_ui.refresh()

    ui.timer(interval=USERINFO_UPDATE_INTERVAL, callback=_user_info_ui_load, immediate=True)

    async def _user_info_ui_relogin() -> None:
        """Relogin to the platform."""
        nonlocal user_info
        user_info = None
        await ui.context.client.connected()
        app.storage.tab["user_info"] = user_info
        _user_info_ui.refresh()
        with contextlib.suppress(Exception):
            user_info = await run.cpu_bound(PlatformService.get_user_info, relogin=True)
            app.storage.tab["user_info"] = user_info
        _user_info_ui.refresh()

    @ui.refreshable
    def health_icon() -> None:
        if launchpad_healthy:
            ui.icon("settings", color="positive")
        elif launchpad_healthy is not None:
            ui.icon("settings", color="negative")

    @ui.refreshable
    def health_link() -> None:
        with (
            ui.link(target="/system").style(
                "background-color: white; text-decoration: none; color: black; padding-left: 10px"
            ),
            ui.row().classes("items-center"),
        ):
            ui.tooltip("Check Launchpad Status")
            if launchpad_healthy:
                ui.icon("check_circle", color="positive")
                ui.label("Launchpad is healthy")
            elif launchpad_healthy is not None:
                ui.icon("error", color="negative")
                ui.label("Launchpad is unhealthy")
            else:
                ui.spinner()

    async def _health_load_and_render() -> None:
        nonlocal launchpad_healthy
        with contextlib.suppress(Exception):
            launchpad_healthy = bool(await run.cpu_bound(SystemService.health_static))
        health_icon.refresh()
        health_link.refresh()

    def _update_health() -> None:
        background_tasks.create_lazy(
            coroutine=_health_load_and_render(),
            name="_health_load_and_render",
        )
        ui.run_javascript("document.getElementById('betterstack').src = document.getElementById('betterstack').src;")

    ui.timer(interval=HEALTH_UPDATE_INTERVAL, callback=_update_health, immediate=True)

    # Set background color based on dark mode
    ui.query("body").classes(
        replace="bg-aignostics-light dark:bg-aignostics-dark"
    )  # https://github.com/zauberzeug/nicegui/pull/448#issuecomment-1492442558

    # Create right_drawer reference before using it
    right_drawer = ui.right_drawer(fixed=True)
    right_drawer.hide()  # Hide by default

    with ui.header(elevated=True).classes("items-center justify-between"):
        with ui.link(target="/"):
            ui.image("/assets/logo.png").style("width: 110px; margin-left: 10px")
            ui.tooltip("Go to start page")
        ui.space()
        if navigation_icon is not None:
            if navigation_icon_tooltip:
                with ui.icon(navigation_icon, color=navigation_icon_color).classes("text-4xl"):
                    ui.tooltip(navigation_icon_tooltip)
            else:
                ui.icon(navigation_icon, color=navigation_icon_color).classes("text-4xl")
        ui.label(navigation_title).classes("text-xl font-bold")
        ui.space()

        dark = ui.dark_mode(app.storage.general.get("dark_mode", False))

        # Fix the dark mode toggle button callback
        def toggle_dark_mode() -> None:
            app.storage.general["dark_mode"] = not app.storage.general.get("dark_mode", False)
            dark.toggle()
            if dark.value:
                ui.query("body").classes(replace="bg-aignostics-dark")
            else:
                ui.query("body").classes(replace="bg-aignostics-light")

        ui.button(
            on_click=toggle_dark_mode,
            icon="dark_mode",
        ).set_visibility(False)

        _user_info_ui()

        with ui.button(icon="folder_special", on_click=lambda _: open_user_data_directory()).props(
            "flat color=purple-400"
        ):
            ui.tooltip("Open data directory of Launchpad")

        with ui.button(on_click=lambda _: right_drawer.toggle(), icon="menu").props(FLAT_COLOR_WHITE):
            ui.tooltip("Open menu")

    if left_sidebar:
        with ui.left_drawer(top_corner=True, bottom_corner=True, elevated=True).props("breakpoint=0"):
            yield
    else:
        yield

    # Populate the right_drawer we created earlier
    with right_drawer, ui.column(align_items="stretch").classes("h-full"):  # noqa: PLR1702
        with ui.list():
            with ui.item(on_click=lambda _: ui.navigate.to("/")).props("clickable"):
                with ui.item_section().props("avatar"):
                    ui.icon("biotech", color="primary")
                with ui.item_section():
                    ui.label("Run Applications").tailwind.font_weight(
                        "bold" if context.client.page.path == "/" else "normal"
                    )
            with ui.item(on_click=lambda _: ui.navigate.to("/dataset/idc")).props("clickable"):
                with ui.item_section().props("avatar"):
                    ui.icon("image", color="primary")
                with ui.item_section():
                    ui.label("Download Datasets").tailwind.font_weight(
                        "bold" if context.client.page.path == "/dataset/idc" else "normal"
                    )
        ui.space()
        with ui.list():
            if find_spec("ijson"):  # noqa: SIM102
                if not (platform.system() == "Linux" and platform.machine() == "aarch64"):
                    with ui.item(on_click=lambda _: ui.navigate.to("/qupath")).props("clickable"):
                        with ui.item_section().props("avatar"):
                            ui.icon("visibility", color="primary")
                        with ui.item_section():
                            ui.label("QuPath Extension").tailwind.font_weight(
                                "bold" if context.client.page.path == "/qupath" else "normal"
                            )
            if find_spec("marimo"):
                with ui.item(on_click=lambda _: ui.navigate.to("/notebook")).props("clickable"):
                    with ui.item_section().props("avatar"):
                        ui.icon("difference", color="primary")
                    with ui.item_section():
                        ui.label("Marimo Extension").tailwind.font_weight(
                            "bold" if context.client.page.path == "/notebook" else "normal"
                        )
            with ui.item(on_click=lambda _: ui.navigate.to("/bucket")).props("clickable"):
                with ui.item_section().props("avatar"):
                    ui.icon("cloud", color="primary")
                with ui.item_section():
                    ui.label("Manage Cloud Bucket").tailwind.font_weight(
                        "bold" if context.client.page.path == "/bucket" else "normal"
                    )
            with ui.item(on_click=lambda _: ui.navigate.to("/system")).props("clickable"):
                with ui.item_section().props("avatar"):
                    health_icon()
                with ui.item_section():
                    ui.label("Info and Settings").tailwind.font_weight(
                        "bold" if context.client.page.path == "/system" else "normal"
                    )
            with ui.item().props("clickable"):
                with ui.item_section().props("avatar"):
                    ui.icon("domain", color="primary")
                with ui.item_section():
                    ui.link("Go to Console", "https://platform.aignostics.com", new_tab=True).mark("LINK_PLATFORM")
            with ui.item().props("clickable"):
                with ui.item_section().props("avatar"):
                    ui.icon("local_library", color="primary")
                with ui.item_section():
                    ui.link("Read The Docs", "https://aignostics.readthedocs.org/", new_tab=True).mark(
                        "LINK_DOCUMENTATION"
                    )
            with ui.item().props("clickable"):
                with ui.item_section().props("avatar"):
                    ui.icon("help", color="primary")
                with ui.item_section():
                    ui.link("Get Support", "https://platform.aignostics.com/support", new_tab=True).mark(
                        "LINK_DOCUMENTATION"
                    )
            with ui.item().props("clickable"):
                with ui.item_section().props("avatar"):
                    ui.icon("check_circle", color="primary")
                with ui.item_section():
                    ui.link("Check Platform Status", "https://status.aignostics.com", new_tab=True).mark(
                        "LINK_DOCUMENTATION"
                    )
            with ui.item().props("clickable"):
                with ui.item_section().props("avatar"):
                    ui.icon("handshake", color="primary")
                with ui.item_section():
                    ui.link(
                        "Attributions", "https://aignostics.readthedocs.io/en/latest/attributions.html", new_tab=True
                    ).mark("LINK_ATTRIBUTIONS")
            if app.native.main_window:
                ui.separator()
                with ui.item(on_click=app.shutdown).props("clickable"):
                    with ui.item_section().props("avatar"):
                        ui.icon("logout", color="primary")
                    with ui.item_section():
                        ui.label("Quit Launcher")
    with (
        ui.footer().style("padding-top:0px; padding-left: 0px; height: 30px; background-color: white"),
        ui.row(align_items="center").classes("justify-start w-full"),
    ):
        health_link()
        with ui.row().style("padding: 0"):
            ui.html(
                '<iframe id="betterstack" src="https://status.aignostics.com/badge?theme=dark" '
                'width="250" height="30" frameborder="0" scrolling="no" '
                'style="color-scheme: dark"></iframe>'
            ).style("margin-left: 0px;")
            ui.tooltip("Check Platform Status")
        ui.space()
        with ui.row():
            ui.html(
                '🔬<a style="color: black; text-decoration: underline" target="_blank" href="https://github.com/aignostics/python-sdk/">'
                f"Aignostics Python SDK v{__version__}</a>"
                ' - built with love in <a style="color: black; text-decoration: underline" target="_blank"'
                ' href="https://www.aignostics.com/company/about">Berlin</A> 🐻'
            ).style("color: black")
            ui.tooltip("Visit GitHub repository of Aignostics Python SDK")



"""CLI (Command Line Interface) of Aignostics Python SDK."""

import sys
from importlib.util import find_spec
from pathlib import Path

import typer

from .constants import MODULES_TO_INSTRUMENT, NOTEBOOK_DEFAULT
from .utils import __is_running_in_container__, __version__, boot, console, get_logger, prepare_cli

boot(MODULES_TO_INSTRUMENT)
logger = get_logger(__name__)

cli = typer.Typer(
    help="Command Line Interface (CLI) of Aignostics Python SDK providing access to Aignostics Platform.",
    no_args_is_help=True,
)

if find_spec("nicegui") and find_spec("webview") and not __is_running_in_container__:

    @cli.command()
    def launchpad() -> None:
        """Open Aignostics Launchpad, the graphical user interface of the Aignostics Platform."""
        from .utils import gui_run  # noqa: PLC0415

        gui_run(native=True, with_api=False, title="Aignostics Launchpad", icon="🔬")


if find_spec("marimo"):
    from typing import Annotated

    from .utils import create_marimo_app

    @cli.command()
    def notebook(
        host: Annotated[str, typer.Option(help="Host to bind the server to")] = "127.0.0.1",
        port: Annotated[int, typer.Option(help="Port to bind the server to")] = 8001,
        notebook: Annotated[
            Path,
            typer.Argument(
                help="Path to the notebook file to run. If not provided, a default notebook will be used.",
                exists=True,
                file_okay=True,
                dir_okay=False,
                readable=True,
            ),
        ] = NOTEBOOK_DEFAULT,
        override_if_exists: Annotated[
            bool,
            typer.Option(
                help="Override the notebook in the user data directory if it already exists.",
            ),
        ] = False,
    ) -> None:
        """Run Python notebook server based on Marimo."""
        import uvicorn  # noqa: PLC0415

        console.print(f"Starting Python notebook server at http://{host}:{port}")
        uvicorn.run(create_marimo_app(notebook=notebook, override_if_exists=override_if_exists), host=host, port=port)


prepare_cli(cli, f"🔬 Aignostics Python SDK v{__version__} - built with love in Berlin 🐻")

if __name__ == "__main__":  # pragma: no cover
    try:
        cli()
    except Exception as e:  # noqa: BLE001
        message = f"An error occurred while running the CLI: {e!s}"
        logger.critical(message)
        console.print(message, style="error")
        sys.exit(1)



"""Aignostics Launchpad launcher for pyinstaller."""

import os
from multiprocessing import freeze_support

freeze_support()

os.environ["LOGFIRE_PYDANTIC_RECORD"] = "off"

from aignostics.constants import MODULES_TO_INSTRUMENT  # noqa: E402
from aignostics.utils import boot, get_logger, gui_run  # noqa: E402

boot(MODULES_TO_INSTRUMENT)
logger = get_logger(__name__)

gui_run(native=True, with_api=False, title="Aignostics Launchpad", icon="🔬")



"""Settings of the Python SDK."""

import logging
import os
from pathlib import Path
from typing import Annotated, TypeVar
from urllib.parse import urlparse

import appdirs
from pydantic import (
    Field,
    FieldSerializationInfo,
    PlainSerializer,
    SecretStr,
    computed_field,
    field_serializer,
    model_validator,
)
from pydantic_settings import BaseSettings, SettingsConfigDict

from aignostics.utils import OpaqueSettings, __project_name__, load_settings

from ._constants import (
    API_ROOT_PRODUCTION,
    AUDIENCE_DEV,
    AUDIENCE_PRODUCTION,
    AUDIENCE_STAGING,
    AUTHORIZATION_BASE_URL_DEV,
    AUTHORIZATION_BASE_URL_PRODUCTION,
    AUTHORIZATION_BASE_URL_STAGING,
    DEVICE_URL_DEV,
    DEVICE_URL_PRODUCTION,
    DEVICE_URL_STAGING,
    JWS_JSON_URL_DEV,
    JWS_JSON_URL_PRODUCTION,
    JWS_JSON_URL_STAGING,
    REDIRECT_URI_DEV,
    REDIRECT_URI_PRODUCTION,
    REDIRECT_URI_STAGING,
    TOKEN_URL_DEV,
    TOKEN_URL_PRODUCTION,
    TOKEN_URL_STAGING,
)
from ._messages import UNKNOWN_ENDPOINT_URL

logger = logging.getLogger(__name__)

T = TypeVar("T", bound=BaseSettings)


class Settings(OpaqueSettings):
    """Configuration settings for the Aignostics SDK.

    This class handles configuration settings loaded from environment variables,
    configuration files, or default values. It manages authentication endpoints,
    client credentials, token storage, and other SDK behaviors.

    Attributes:
        api_root (str): Base URL of the Aignostics API.
        audience (str): OAuth audience claim.
        authorization_backoff_seconds (int): Backoff time for authorization retries in seconds.
        authorization_base_url (str): Authorization endpoint for OAuth flows.
        cache_dir (str): Directory for caching tokens and other data.
        client_id_device (SecretStr): Client ID for device authorization flow.
        client_id_interactive (SecretStr): Client ID for interactive authorization flow.
        device_url (str): Device authorization endpoint for device flow.
        jws_json_url (str): URL for JWS key set.
        redirect_uri (str): Redirect URI for OAuth authorization code flow.
        refresh_token (SecretStr | None): OAuth refresh token if available.
        request_timeout_seconds (int): Timeout for API requests in seconds.
        scope (str): OAuth scopes required by the SDK.
        scope_elements (list[str]): OAuth scopes split into individual elements.
        token_file (Path): Path to the token storage file.
        token_url (str): Token endpoint for OAuth flows.
    """

    model_config = SettingsConfigDict(
        env_prefix=f"{__project_name__.upper()}_",
        env_file=(
            os.getenv(f"{__project_name__.upper()}_ENV_FILE", Path.home() / f".{__project_name__}/.env"),
            Path(".env"),
        ),
        env_file_encoding="utf-8",
        extra="ignore",
    )

    client_id_device: Annotated[
        SecretStr,
        PlainSerializer(
            func=OpaqueSettings.serialize_sensitive_info, return_type=str, when_used="always"
        ),  # allow to unhide sensitive info from CLI or if user presents valid token via API
        Field(description="OAuth Client ID Interactive"),
    ]
    client_id_interactive: Annotated[
        SecretStr,
        PlainSerializer(
            func=OpaqueSettings.serialize_sensitive_info, return_type=str, when_used="always"
        ),  # allow to unhide sensitive info from CLI or if user presents valid token via API
        Field(description="OAuth Client ID Interactive"),
    ]
    api_root: Annotated[
        str,
        Field(description="URL of the API root", default=API_ROOT_PRODUCTION),
    ]

    scope: str = "offline_access, profile, email, openid"

    @computed_field  # type: ignore[prop-decorator]
    @property
    def scope_elements(self) -> list[str]:
        """Get the OAuth scope elements as a list.

        Splits the scope string by comma and strips whitespace from each element.

        Returns:
            list[str]: List of individual scope elements.
        """
        if not self.scope:
            return []
        return [element.strip() for element in self.scope.split(",")]

    audience: str
    authorization_base_url: str

    @computed_field  # type: ignore[prop-decorator]
    @property
    def issuer(self) -> str:
        """Get the issuer URL based on the authorization base root.

        Extracts the scheme and domain from the authorization base URL to create
        a failsafe issuer URL in the format scheme://domain/

        Returns:
            str: Issuer URL in the format scheme://domain/
        """
        try:
            parsed = urlparse(self.authorization_base_url)
            if parsed.scheme and parsed.netloc:
                return f"{parsed.scheme}://{parsed.netloc}/"
            # Fallback to original logic if URL parsing fails
            logger.warning(
                "Failed to parse authorization_base_url '%s', falling back to rsplit method",
                self.authorization_base_url,
            )
            return self.authorization_base_url.rsplit("/", 1)[0] + "/"
        except (ValueError, AttributeError):
            # Ultimate fallback if everything fails
            logger.exception(
                "Error parsing authorization_base_url '%s', falling back to rsplit method",
                self.authorization_base_url,
            )
            return self.authorization_base_url.rsplit("/", 1)[0] + "/"

    token_url: str
    redirect_uri: str
    device_url: str
    jws_json_url: str

    refresh_token: Annotated[
        SecretStr | None,
        PlainSerializer(
            func=OpaqueSettings.serialize_sensitive_info, return_type=str, when_used="always"
        ),  # allow to unhide sensitive info from CLI or if user presents valid token via API
        Field(description="Refresh token for OAuth authentication", min_length=10, max_length=1000, default=None),
    ] = None

    cache_dir: str = appdirs.user_cache_dir(__project_name__)

    @computed_field  # type: ignore[prop-decorator]
    @property
    def token_file(self) -> Path:
        """Get the path to the token file.

        Returns:
            Path: The path to the file where the authentication token is stored.
        """
        return Path(self.cache_dir) / ".token"

    @field_serializer("token_file")
    def serialize_token_file(self, token_file: Path, _info: FieldSerializationInfo) -> str:  # noqa: PLR6301
        return str(token_file.resolve())

    request_timeout_seconds: int = 30
    authorization_backoff_seconds: int = 3

    @model_validator(mode="before")
    def pre_init(cls, values: dict) -> dict:  # type: ignore[type-arg] # noqa: N805
        """Initialize auth-related fields based on the API root.

        This validator sets the appropriate authentication URLs and parameters
        based on the target environment (production, staging, or development).
        If auth-related fields are already provided, they will not be overridden.

        Args:
            values: The input data dictionary to validate.

        Returns:
            The updated values dictionary with all environment-specific fields populated.

        Raises:
            ValueError: If the API root URL is not recognized and auth fields are missing.
        """
        # See https://github.com/pydantic/pydantic/issues/9789
        api_root = values.get("api_root", API_ROOT_PRODUCTION)

        # Check if all required auth fields are already provided
        auth_fields = ["audience", "authorization_base_url", "token_url", "redirect_uri", "device_url", "jws_json_url"]
        all_auth_fields_provided = all(field in values for field in auth_fields)

        # If all auth fields are provided, don't override them
        if all_auth_fields_provided:
            return values

        match api_root:
            case "https://platform.aignostics.com":
                values["audience"] = AUDIENCE_PRODUCTION
                values["authorization_base_url"] = AUTHORIZATION_BASE_URL_PRODUCTION
                values["token_url"] = TOKEN_URL_PRODUCTION
                values["redirect_uri"] = REDIRECT_URI_PRODUCTION
                values["device_url"] = DEVICE_URL_PRODUCTION
                values["jws_json_url"] = JWS_JSON_URL_PRODUCTION
            case "https://platform-staging.aignostics.com":
                values["audience"] = AUDIENCE_STAGING
                values["authorization_base_url"] = AUTHORIZATION_BASE_URL_STAGING
                values["token_url"] = TOKEN_URL_STAGING
                values["redirect_uri"] = REDIRECT_URI_STAGING
                values["device_url"] = DEVICE_URL_STAGING
                values["jws_json_url"] = JWS_JSON_URL_STAGING
            case "https://platform-dev.aignostics.com":
                values["audience"] = AUDIENCE_DEV
                values["authorization_base_url"] = AUTHORIZATION_BASE_URL_DEV
                values["token_url"] = TOKEN_URL_DEV
                values["redirect_uri"] = REDIRECT_URI_DEV
                values["device_url"] = DEVICE_URL_DEV
                values["jws_json_url"] = JWS_JSON_URL_DEV
            case _:
                raise ValueError(UNKNOWN_ENDPOINT_URL)

        return values


__cached_settings: Settings | None = None


def settings() -> Settings:
    """Lazy load authentication settings from the environment or a file.

    * Given we use Pydantic Settings, validation is done automatically.
    * We only load and validate if we actually need the settings,
        thereby not killing the client on other actions.
    * If the settings have already been loaded, return the cached instance.

    Returns:
        AuthenticationSettings: The loaded authentication settings.
    """
    global __cached_settings  # noqa: PLW0603
    if __cached_settings is None:
        __cached_settings = load_settings(Settings)  # pyright: ignore[reportCallIssue]
    return __cached_settings



"""Base class for services."""

from abc import ABC, abstractmethod
from typing import Any, TypeVar

from pydantic_settings import BaseSettings

from ._health import Health
from ._settings import load_settings

T = TypeVar("T", bound=BaseSettings)


class BaseService(ABC):
    """Base class for services."""

    _settings: BaseSettings

    def __init__(self, settings_class: type[T] | None = None) -> None:
        """
        Initialize service with optional settings.

        Args:
            settings_class: Optional settings class to load configuration.
        """
        if settings_class is not None:
            self._settings = load_settings(settings_class)

    def key(self) -> str:
        """Return the module name of the instance."""
        return self.__module__.split(".")[-2]

    @abstractmethod
    def health(self) -> Health:
        """Get health of this service. Override in subclass.

        Returns:
            Health: Health status of the service.
        """

    @abstractmethod
    def info(self, mask_secrets: bool = True) -> dict[str, Any]:
        """Get info of this service. Override in subclass.

        Args:
            mask_secrets: Whether to mask sensitive information in the output.

        Returns:
            dict[str, Any]: Information about the service.
        """



"""Settings of the bucket module."""

from enum import StrEnum
from typing import Annotated

from pydantic import Field, PlainSerializer, SecretStr
from pydantic_settings import SettingsConfigDict

from ..utils import OpaqueSettings, __env_file__, __project_name__  # noqa: TID252


class BucketProtocol(StrEnum):
    GS = "gs"
    S3 = "s3"


class Settings(OpaqueSettings):
    """Settings."""

    model_config = SettingsConfigDict(
        env_prefix=f"{__project_name__.upper()}_BUCKET_",
        extra="ignore",
        env_file=__env_file__,
        env_file_encoding="utf-8",
    )

    protocol: Annotated[
        BucketProtocol,
        Field(
            description=("Protocol to access the cloud bucket, default is gs"),
            default="gs",
        ),
    ]

    name: Annotated[
        str,
        Field(description=("Name of the cloud bucket")),
    ]

    region_name: Annotated[
        str,
        Field(description=("Region of the cloud bucket"), default="EUROPE-WEST3"),
    ]

    hmac_access_key_id: Annotated[
        SecretStr,
        PlainSerializer(func=OpaqueSettings.serialize_sensitive_info, return_type=str, when_used="always"),
        Field(description=("HMAC access key ID of the cloud bucket")),
    ]

    hmac_secret_access_key: Annotated[
        SecretStr,
        PlainSerializer(func=OpaqueSettings.serialize_sensitive_info, return_type=str, when_used="always"),
        Field(description=("HMAC secret access key of the cloud bucket")),
    ]

    upload_signed_url_expiration_seconds: Annotated[
        int,
        Field(
            description=("Expiration time for signed URLs created and used by the Python SDK to upload to the bucket."),
            default=2
            * 60
            * 60,  # The Python SDK creates the signed upload URLs immediately before uploading, so 2h is sufficient.
            ge=60,  # Minimum expiration time is 60 seconds, see https://docs.aws.amazon.com/AmazonS3/latest/userguide/using-presigned-url.html
            le=7
            * 24
            * 60
            * 60,  # Limit to 7 days as this is the max e.g. at AWS, see https://docs.aws.amazon.com/AmazonS3/latest/userguide/using-presigned-url.html
        ),
    ]

    download_signed_url_expiration_seconds: Annotated[
        int,
        Field(
            description=(
                "Expiration time of the signed URLs provided by the Python SDK to the platform on submitting "
                "application runs."
            ),
            default=7 * 24 * 60 * 60,  # The platform queues application runs, so we set a gracious default of 7d.
            ge=60,  # Minimum expiration time is 60 seconds, see https://docs.aws.amazon.com/AmazonS3/latest/userguide/using-presigned-url.html
            le=7
            * 24
            * 60
            * 60,  # Limit to 7 days as this is the max e.g. at AWS, see https://docs.aws.amazon.com/AmazonS3/latest/userguide/using-presigned-url.html
        ),
    ]



"""CLI of QuPath module."""

import platform
import sys
from pathlib import Path
from typing import Annotated

import typer
from rich.table import Table

from aignostics.utils import console, get_logger

from ._service import QUPATH_VERSION, Service

logger = get_logger(__name__)


cli = typer.Typer(
    name="qupath",
    help="Interact with QuPath application.",
)


@cli.command()
def install(
    version: Annotated[
        str,
        typer.Option(
            help="Version of QuPath to install. Do not change this unless you know what you are doing.",
        ),
    ] = QUPATH_VERSION,
    path: Annotated[
        Path,
        typer.Option(
            help="Path to install QuPath to. If not specified, the default installation path will be used."
            "Do not change this unless you know what you are doing.",
            exists=False,
            file_okay=False,
            dir_okay=True,
            writable=True,
            readable=True,
            resolve_path=True,
        ),
    ] = Service.get_installation_path(),  # noqa: B008
    reinstall: Annotated[
        bool,
        typer.Option(
            help="Reinstall QuPath even if it is already installed. This will overwrite the existing installation.",
        ),
    ] = True,
    platform_system: Annotated[
        str,
        typer.Option(help="Override the system to assume for the installation. This is useful for testing purposes."),
    ] = platform.system(),
    platform_machine: Annotated[
        str,
        typer.Option(
            help="Override the machine architecture to assume for the installation. "
            "This is useful for testing purposes.",
        ),
    ] = platform.machine(),
) -> None:
    """Install QuPath application."""
    from rich.progress import (  # noqa: PLC0415
        BarColumn,
        FileSizeColumn,
        Progress,
        TaskProgressColumn,
        TextColumn,
        TimeRemainingColumn,
        TotalFileSizeColumn,
        TransferSpeedColumn,
    )

    try:
        console.print(f"Installing QuPath version {version} to {path}...")
        with Progress(
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            TaskProgressColumn(),
            TimeRemainingColumn(),
            FileSizeColumn(),
            TotalFileSizeColumn(),
            TransferSpeedColumn(),
            TextColumn("[progress.description]{task.fields[extra_description]}"),
        ) as progress:
            download_task = progress.add_task("Downloading", total=None, extra_description="")
            extract_task = progress.add_task("Extracting", total=None, extra_description="")

            def download_progress(filepath: Path, filesize: int, chunksize: int) -> None:
                progress.update(
                    download_task,
                    total=filesize,
                    advance=chunksize,
                    extra_description=filepath.name,
                )

            def extract_progress(application_path: Path, application_size: int) -> None:
                progress.update(
                    extract_task,
                    total=application_size,
                    completed=application_size,
                    extra_description=application_path,
                )

            application_path = Service().install_qupath(
                version=version,
                path=path,
                reinstall=reinstall,
                platform_system=platform_system,
                platform_machine=platform_machine,
                download_progress=download_progress,
                extract_progress=extract_progress,
            )

        console.print(f"QuPath v{version} installed successfully at '{application_path!s}'", style="success")
    except Exception as e:
        message = f"Failed to install QuPath version {version} at {path!s}: {e!s}."
        logger.exception(message)
        console.print(message, style="error")
        sys.exit(1)


@cli.command()
def launch(
    project: Annotated[
        Path | None,
        typer.Option(
            help="Path to QuPath project directory.",
            exists=True,
            file_okay=False,
            dir_okay=True,
            writable=True,
            readable=True,
            resolve_path=True,
        ),
    ] = None,
    image: Annotated[
        str | None,
        typer.Option(
            help="Path to image. Must be part of QuPath project",
        ),
    ] = None,
    script: Annotated[
        Path | None,
        typer.Option(
            help="Path to QuPath script to run on launch. Must be part of QuPath project.",
            exists=True,
            file_okay=True,
            dir_okay=False,
            writable=False,
            readable=True,
            resolve_path=True,
        ),
    ] = None,
) -> None:
    """Launch QuPath application."""
    try:
        if not Service().is_qupath_installed():
            console.print("QuPath is not installed. Use 'uvx aignostics qupath install' to install it.")
            sys.exit(2)
        pid = Service.execute_qupath(project=project, image=image, script=script)
        if not pid:
            console.print("QuPath could not be launched.", style="error")
            sys.exit(1)
        message = f"QuPath launched successfully with process id '{pid}'."
        console.print(message, style="success")
    except Exception as e:
        message = f"Failed to launch QuPath: {e!s}."
        logger.exception(message)
        console.print(message, style="error")
        sys.exit(1)


@cli.command()
def processes(
    json: Annotated[
        bool,
        typer.Option(
            "--json",
            "-j",
            help="Output the running QuPath processes as JSON.",
            is_flag=True,
        ),
    ],
) -> None:
    """List running QuPath processes.

    Notice: This will not list processes that are not started from the installation directory.
    """
    try:
        processes = Service.get_qupath_processes()
        if not processes:
            console.print("No running QuPath processes found.", style="warning")
            sys.exit(0)
        if json:
            process_info = [
                {
                    "pid": process.pid,
                    "exe": process.exe() or "N/A",
                    "cwd": process.cwd() or "N/A",
                    "args": " ".join(process.cmdline()[1:]) if len(process.cmdline()) > 1 else "N/A",
                }
                for process in processes
            ]
            console.print_json(data=process_info)
            return
        process_table = Table(title="Running QuPath Processes")
        process_table.add_column("Process ID", justify="right", style="cyan")
        process_table.add_column("Executable", style="green")
        process_table.add_column("Working Directory", style="yellow")
        process_table.add_column("Arguments", style="blue")
        for process in processes:
            process_table.add_row(
                str(process.pid),
                process.exe() or "N/A",
                process.cwd() or "N/A",
                " ".join(process.cmdline()[1:]) if len(process.cmdline()) > 1 else "N/A",
            )
        console.print(process_table)
    except Exception as e:
        message = f"Failed to determine running QuPath processes: {e!s}."
        logger.exception(message)
        console.print(message, style="error")
        sys.exit(1)


@cli.command()
def terminate() -> None:
    """Terminate running QuPath processes.

    Notice: This will not terminate processes that are not started from the installation directory.
    """
    try:
        terminated_count = Service.terminate_qupath_processes()
        if terminated_count == 0:
            console.print("No running QuPath processes found to terminate.", style="warning")
            sys.exit(2)
        else:
            console.print(f"Terminated {terminated_count} running QuPath processes.", style="success")
    except Exception as e:
        message = f"Failed to determine running QuPath processes: {e!s}."
        logger.exception(message)
        console.print(message, style="error")
        sys.exit(1)


@cli.command()
def uninstall(
    version: Annotated[
        str | None,
        typer.Option(
            help="Version of QuPath to install. If not specified, all versions will be uninstalled.",
        ),
    ] = None,
    path: Annotated[
        Path,
        typer.Option(
            help="Path to install QuPath to. If not specified, the default installation path will be used."
            "Do not change this unless you know what you are doing.",
            exists=False,
            file_okay=False,
            dir_okay=True,
            writable=True,
            readable=True,
            resolve_path=True,
        ),
    ] = Service.get_installation_path(),  # noqa: B008
    platform_system: Annotated[
        str,
        typer.Option(help="Override the system to assume for the installation. This is useful for testing purposes."),
    ] = platform.system(),
    platform_machine: Annotated[
        str,
        typer.Option(
            help="Override the machine architecture to assume for the installation. "
            "This is useful for testing purposes.",
        ),
    ] = platform.machine(),
) -> None:
    """Uninstall QuPath application."""
    try:
        uninstalled = Service().uninstall_qupath(version, path, platform_system, platform_machine)
        if not uninstalled:
            console.print(f"QuPath not installed at {path!s}.", style="warning")
            sys.exit(2)
        console.print("QuPath uninstalled successfully.", style="success")
    except Exception as e:
        message = f"Failed to uninstall QuPath version {version} at {path!s}: {e!s}."
        logger.exception(message)
        console.print(message, style="error")
        sys.exit(1)


@cli.command()
def add(
    project: Annotated[
        Path,
        typer.Argument(
            help="Path to QuPath project directory. Will be created if it does not exist.",
            exists=False,
            file_okay=False,
            dir_okay=True,
            writable=True,
            readable=True,
            resolve_path=True,
        ),
    ],
    path: Annotated[
        list[Path],
        typer.Argument(
            help="One or multiple paths. A path can point to an individual image or folder."
            "In case of a folder, all images within will be added for supported image types.",
            exists=True,
            file_okay=True,
            dir_okay=True,
            writable=False,
            readable=True,
            resolve_path=True,
        ),
    ],
) -> None:
    """Add image(s) to QuPath project. Creates project if it does not exist."""
    try:
        count = Service().add(
            project=project,
            paths=path,
        )
        console.print(f"Added '{count}' images to project '{project}'.", style="success")
    except Exception as e:
        message = f"Failed to add images to project: {e!s}."
        logger.exception(message)
        console.print(message, style="error")
        sys.exit(1)


@cli.command()
def annotate(
    project: Annotated[
        Path,
        typer.Argument(
            help="Path to QuPath project directory. Will be created if it does not exist.",
            exists=False,
            file_okay=False,
            dir_okay=True,
            writable=True,
            readable=True,
            resolve_path=True,
        ),
    ],
    image: Annotated[
        Path,
        typer.Argument(
            help="Path to image to annotate. If the image is not part of the project, it will be added.",
            exists=True,
            file_okay=True,
            dir_okay=False,
            writable=False,
            readable=True,
            resolve_path=True,
        ),
    ],
    annotations: Annotated[
        Path,
        typer.Argument(
            help="Path to polygons file to import. The file must be a compatible GeoJSON file.",
            exists=True,
            file_okay=True,
            dir_okay=False,
            writable=False,
            readable=True,
            resolve_path=True,
        ),
    ],
) -> None:
    """Add image(s) to QuPath project. Creates project if it does not exist."""
    try:
        annotation_count = Service().annotate(project=project, image=image, annotations=annotations)
        console.print(
            f"Added {annotation_count} annotations to {image} in {project}.",
            style="success",
        )
    except Exception as e:
        message = f"Failed to add images to project: {e!s}."
        logger.exception(message)
        console.print(message, style="error")
        sys.exit(1)


@cli.command()
def inspect(
    project: Annotated[
        Path,
        typer.Argument(
            help="Path to QuPath project directory.",
            exists=True,
            file_okay=False,
            dir_okay=True,
            writable=False,
            readable=True,
            resolve_path=True,
        ),
    ],
) -> None:
    """Inspect project."""
    try:
        info = Service().inspect(project=project)
        console.print_json(data=info.model_dump())
    except Exception as e:
        message = f"Failed to read project: {e!s}."
        logger.exception(message)
        console.print(message, style="error")
        sys.exit(1)


@cli.command(name="run-script")
def run_script(
    script: Annotated[
        Path,
        typer.Argument(
            help="Path to the Groovy script file to execute.",
            exists=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
            resolve_path=True,
        ),
    ],
    project: Annotated[
        Path | None,
        typer.Option(
            "--project",
            "-p",
            help="Path to the QuPath project directory.",
            exists=False,
            file_okay=False,
            dir_okay=True,
            writable=True,
            readable=True,
            resolve_path=True,
        ),
    ] = None,
    image: Annotated[
        str | None,
        typer.Option(
            "--image",
            "-i",
            help="Name of the image in the project or path to image file.",
        ),
    ] = None,
    args: Annotated[
        list[str] | None,
        typer.Option(
            "--args",
            "-a",
            help="Arguments to pass to the script. Can be specified multiple times.",
        ),
    ] = None,
) -> None:
    """Run a QuPath Groovy script with optional arguments."""
    try:
        if not Service().is_qupath_installed():
            console.print("QuPath is not installed. Use 'uvx aignostics qupath install' to install it.")
            sys.exit(2)

        # Validate script file exists
        if not script.is_file():
            console.print(f"Script file not found: {script}", style="error")
            sys.exit(1)

        pid = Service.execute_qupath(
            quiet=True,
            project=project,
            image=image,
            script=script,
            script_args=args,
        )

        if not pid:
            console.print("QuPath script could not be executed.", style="error")
            sys.exit(1)

        message = f"QuPath script executed successfully with process id '{pid}'."
        console.print(message, style="success")

        if args:
            console.print(f"Script arguments: {' '.join(args)}", style="info")

    except Exception as e:
        message = f"Failed to run QuPath script: {e!s}."
        logger.exception(message)
        console.print(message, style="error")
        sys.exit(1)



"""Homepage (index) of GUI."""

from pathlib import Path

from aignostics.gui import frame
from aignostics.utils import BaseService, locate_subclasses

from ..utils import BasePageBuilder  # noqa: TID252
from ._service import Service


class PageBuilder(BasePageBuilder):
    @staticmethod
    def register_pages() -> None:  # noqa: PLR0915
        from nicegui import app, run, ui  # noqa: PLC0415

        locate_subclasses(BaseService)  # Ensure settings are loaded
        app.add_static_files("/system_assets", Path(__file__).parent / "assets")

        ui.add_head_html("""
            <style>
                :global(.jse-modal-window.jse-modal-window-jsoneditor)
                {
                    width: 100%;
                    height: 100%;
                    min-height: 900px;
                }
            </style>
        """)

        @ui.page("/system")
        async def page_system() -> None:  # noqa: PLR0915
            """System info and settings page."""
            with frame("Info and Settings", left_sidebar=False):
                pass

            with ui.row().classes("w-full gap-4 flex-nowrap"):
                with ui.column().classes("w-3/5 flex-shrink-0"):
                    with ui.tabs().classes("w-full") as tabs:
                        tab_health = ui.tab("Health")
                        tab_info = ui.tab("Info")
                        tab_settings = ui.tab("Settings")
                    with ui.tab_panels(tabs, value=tab_health).classes("w-full"):
                        with ui.tab_panel(tab_health).classes("min-h-[calc(100vh-12rem)]"):
                            spinner = ui.spinner(size="lg").classes(
                                "absolute top-1/2 left-1/2 transform -translate-x-1/2 -translate-y-1/2 z-10"
                            )
                            properties = {
                                "content": {"json": "Loading ..."},
                                "mode": "tree",
                                "readOnly": True,
                                "mainMenuBar": True,
                                "navigationBar": True,
                                "statusBar": True,
                            }
                            editor = ui.json_editor(properties).style("width: 100%").mark("JSON_EDITOR_HEALTH")
                            editor.set_visibility(False)
                            health = await run.cpu_bound(Service.health_static)
                            if health is None:
                                properties["content"] = {"json": "Health check failed."}  # type: ignore[unreachable]
                            else:
                                properties["content"] = {"json": health.model_dump()}
                            editor.update()
                            editor.run_editor_method(":expand", "path => true")
                            spinner.set_visibility(False)
                            editor.set_visibility(True)
                        with ui.tab_panel(tab_info).classes("min-h-[calc(100vh-12rem)]"):
                            # Mask secrets switch with reload functionality
                            with ui.row().classes("w-full items-center gap-2 mb-4"):
                                mask_secrets_switch = ui.switch(
                                    text="Mask secrets", value=True, on_change=lambda e: load_info(mask_secrets=e.value)
                                )

                            spinner = ui.spinner(size="lg").classes(
                                "absolute top-1/2 left-1/2 transform -translate-x-1/2 -translate-y-1/2 z-10"
                            )
                            properties = {
                                "content": {"json": "Loading ..."},
                                "mode": "tree",
                                "readOnly": True,
                                "mainMenuBar": True,
                                "navigationBar": True,
                                "statusBar": True,
                            }
                            editor = ui.json_editor(properties).style("width: 100%").mark("JSON_EDITOR_INFO")

                            async def load_info(mask_secrets: bool = True) -> None:
                                """Load system info with current mask_secrets setting."""
                                editor.set_visibility(False)
                                spinner.set_visibility(True)
                                mask_secrets_switch.set_visibility(False)
                                info = await run.cpu_bound(
                                    Service.info, include_environ=True, mask_secrets=mask_secrets
                                )
                                if info is None:
                                    properties["content"] = {"json": "Info retrieval failed."}  # type: ignore[unreachable]
                                else:
                                    properties["content"] = {"json": info}
                                editor.update()
                                editor.run_editor_method(":expand", "path => true")
                                spinner.set_visibility(False)
                                editor.set_visibility(True)
                                mask_secrets_switch.set_visibility(True)

                            # Initial load
                            editor.set_visibility(False)
                            await load_info()
                        with (
                            ui.tab_panel(tab_settings),
                            ui.card().classes("w-full"),
                            ui.row().classes("items-center justify-between"),
                        ):
                            ui.switch(
                                value=Service.remote_diagnostics_enabled(),
                                on_change=lambda e: (
                                    Service.remote_diagnostics_enable()
                                    if e.value
                                    else Service.remote_diagnostics_disable(),
                                    ui.notify("Restart the app to apply changes.", color="warning"),  # type: ignore[func-returns-value]
                                    None,
                                )[0],
                            )
                            ui.label("Remote Diagnostics")
                with ui.column().classes("w-2/5 flex-shrink-0 flex items-center justify-start mt-[200px]"):
                    ui.html(
                        '<dotlottie-player src="/system_assets/system.lottie" '
                        'background="transparent" speed="1" style="width: 300px; height: 300px" '
                        'direction="1" playMode="normal" loop autoplay></dotlottie-player>'
                    )



"""System CLI commands."""

import json
import sys
from enum import StrEnum
from importlib.util import find_spec
from typing import Annotated

import typer
import yaml

from ..constants import API_VERSIONS  # noqa: TID252
from ..utils import console, get_logger  # noqa: TID252
from ._service import Service

logger = get_logger(__name__)

cli = typer.Typer(name="system", help="Determine health, info and further utillities.")

_service = Service()

HTTP_PROXY_DEFAULT_HOST = "proxy.charite.de"
HTTP_PROXY_DEFAULT_PORT = 8080
HTTP_PROXY_DEFAULT_SCHEME = "http"


class OutputFormat(StrEnum):
    """
    Enum representing the supported output formats.

    This enum defines the possible formats for output data:
    - YAML: Output data in YAML format
    - JSON: Output data in JSON format

    Usage:
        format = OutputFormat.YAML
        print(f"Using {format} format")
    """

    YAML = "yaml"
    JSON = "json"


@cli.command()
def health(
    output_format: Annotated[
        OutputFormat, typer.Option(help="Output format", case_sensitive=False)
    ] = OutputFormat.JSON,
) -> None:
    """Determine and print system health.

    Args:
        output_format (OutputFormat): Output format (JSON or YAML).
    """
    match output_format:
        case OutputFormat.JSON:
            console.print_json(data=_service.health().model_dump())
        case OutputFormat.YAML:
            console.print(
                yaml.dump(data=json.loads(_service.health().model_dump_json()), width=80, default_flow_style=False),
                end="",
            )


@cli.command()
def info(
    include_environ: Annotated[bool, typer.Option(help="Include environment variables")] = False,
    mask_secrets: Annotated[bool, typer.Option(help="Mask secrets")] = True,
    output_format: Annotated[
        OutputFormat, typer.Option(help="Output format", case_sensitive=False)
    ] = OutputFormat.JSON,
) -> None:
    """Determine and print system info.

    Args:
        include_environ (bool): Include environment variables.
        mask_secrets (bool): Mask values for variables identified as secrets.
        output_format (OutputFormat): Output format (JSON or YAML).
    """
    info = _service.info(include_environ=include_environ, mask_secrets=mask_secrets)
    match output_format:
        case OutputFormat.JSON:
            console.print_json(data=info)
        case OutputFormat.YAML:
            console.print(yaml.dump(info, width=80, default_flow_style=False), end="")


if find_spec("nicegui"):
    from ..utils import gui_run  # noqa: TID252

    @cli.command()
    def serve(
        host: Annotated[str, typer.Option(help="Host to bind the server to")] = "127.0.0.1",
        port: Annotated[int, typer.Option(help="Port to bind the server to")] = 8000,
        open_browser: Annotated[bool, typer.Option(help="Open app in browser after starting the server")] = False,
    ) -> None:
        """Start the web server, hosting the graphical web application and/or webservice API.

        Args:
            host (str): Host to bind the server to.
            port (int): Port to bind the server to.
            watch (bool): Enable auto-reload on changes of source code.
            open_browser (bool): Open app in browser after starting the server.
        """
        console.print(f"Starting web application server at http://{host}:{port}")
        gui_run(native=False, host=host, port=port, with_api=False, show=open_browser)


@cli.command()
def openapi(
    api_version: Annotated[
        str, typer.Option(help=f"API Version. Available: {', '.join(API_VERSIONS.keys())}", case_sensitive=False)
    ] = next(iter(API_VERSIONS.keys())),
    output_format: Annotated[
        OutputFormat, typer.Option(help="Output format", case_sensitive=False)
    ] = OutputFormat.JSON,
) -> None:
    """Dump the OpenAPI specification.

    Args:
        api_version (str): API version to dump.
        output_format (OutputFormat): Output format (JSON or YAML).

    Raises:
        typer.Exit: If an invalid API version is provided.
    """
    match api_version:
        case "v1":
            schema = Service.openapi_schema()
        case _:
            available_versions = ", ".join(API_VERSIONS.keys())
            console.print(
                f"[bold red]Error:[/] Invalid API version '{api_version}'. Available versions: {available_versions}"
            )
            raise typer.Exit(code=1)
    match output_format:
        case OutputFormat.JSON:
            console.print_json(data=schema)
        case OutputFormat.YAML:
            console.print(yaml.dump(schema, default_flow_style=False), end="")


@cli.command()
def install() -> None:
    """Complete installation."""
    console.print("Installation complete!")


config_app = typer.Typer()
cli.add_typer(config_app, name="config", help="Configure application settings.")


@config_app.command()
def get(key: Annotated[str, typer.Argument(help="Configuration key to get value for")]) -> None:
    """Set a configuration key to a value."""
    console.print(Service.dotenv_get(key.upper()))


@config_app.command()
def set(  # noqa: A001
    key: Annotated[str, typer.Argument(help="Configuration key to set")],
    value: Annotated[str, typer.Argument(help="Value to set for the configuration key")],
) -> None:
    """Set a configuration key to a value."""
    key = key.upper()
    try:
        Service.dotenv_set(key, value)
    except ValueError as e:
        console.print(f"Invalid configuration: {e!s}", style="error")
        sys.exit(2)
    console.print(f"Configuration '{key}' set to '{value}'.", style="success")


@config_app.command()
def unset(
    key: Annotated[str, typer.Argument(help="Configuration key to unset")],
) -> None:
    """Set a configuration key to a value."""
    key = key.upper()
    Service.dotenv_unset(key)
    console.print(f"Configuration '{key}' unset.", style="success")


@config_app.command()
def remote_diagnostics_enable() -> None:
    """Enable remote diagnostics via Sentry and Logfire. Data stored in EU data centers."""
    try:
        Service.remote_diagnostics_enable()
        console.print("Remote diagnostics enabled.", style="success")
    except ValueError as e:
        console.print(f"Invalid diagnostics configuration: {e!s}", style="error")
        sys.exit(2)


@config_app.command()
def remote_diagnostics_disable() -> None:
    """Disable remote diagnostics."""
    try:
        Service.remote_diagnostics_disable()
        console.print("Remote diagnostics disabled.", style="success")
    except ValueError as e:
        console.print(f"Invalid diagnostics configuration: {e!s}", style="error")
        sys.exit(2)


@config_app.command()
def http_proxy_enable(
    host: Annotated[str, typer.Option(help="Host")] = HTTP_PROXY_DEFAULT_HOST,
    port: Annotated[int, typer.Option(help="Port")] = HTTP_PROXY_DEFAULT_PORT,
    scheme: Annotated[str, typer.Option(help="Scheme")] = HTTP_PROXY_DEFAULT_SCHEME,
    ssl_cert_file: Annotated[str | None, typer.Option(help="SSL certificate file")] = None,
    no_ssl_verify: Annotated[bool, typer.Option(help="Disable SSL verification")] = False,
) -> None:
    """Enable HTTP proxy."""
    try:
        Service().http_proxy_enable(
            host=host, port=port, scheme=scheme, ssl_cert_file=ssl_cert_file, no_ssl_verify=no_ssl_verify
        )
        console.print("HTTP proxy enabled.", style="success")
    except ValueError as e:
        console.print(f"Invalid HTTP proxy configuration: {e!s}", style="error")
        sys.exit(2)


@config_app.command()
def http_proxy_disable() -> None:
    """Disable HTTP proxy."""
    try:
        Service.http_proxy_disable()
        console.print("HTTP proxy disabled.", style="success")
    except ValueError as e:
        console.print(f"Invalid HTTP proxy configuration: {e!s}", style="error")
        sys.exit(2)



import errno
import socket
import time
import typing as t
import webbrowser
from datetime import UTC, datetime, timedelta
from http.server import BaseHTTPRequestHandler, HTTPServer
from pathlib import Path
from urllib import parse
from urllib.error import HTTPError

import jwt
import requests
from auth0.authentication import Users
from pydantic import BaseModel, SecretStr
from requests_oauthlib import OAuth2Session

from aignostics.platform._messages import AUTHENTICATION_FAILED, INVALID_REDIRECT_URI
from aignostics.platform._settings import settings

try:
    import sentry_sdk
except ImportError:
    sentry_sdk = None  # type: ignore[assignment]

CALLBACK_PORT_RETRY_COUNT = 10
CLAIM_ROLE = "https://aignostics-platform-samia/role"


class AuthenticationResult(BaseModel):
    """Represents the result of an OAuth authentication flow."""

    token: str | None = None
    error: str | None = None


def _inform_sentry_about_user(token: str) -> None:
    """Inform Sentry about the authenticated user from the JWT token.

    Args:
        token: The JWT access token to extract user information from.

    Raises:
        RuntimeError: If the token does not contain the 'sub' claim or if verification fails.
    """
    if sentry_sdk is None:
        return  # type: ignore[unreachable]

    def _raise_missing_sub_claim_error() -> None:
        """Raises RuntimeError for missing 'sub' claim in token.

        Raises:
            RuntimeError: If the 'sub' claim is missing from the token.
        """
        message = "Token does not contain 'sub' claim, cannot set Sentry user info."
        raise RuntimeError(message)

    try:
        claims = verify_and_decode_token(token)
        user_id = claims.get("sub", None)
        if user_id is None:
            _raise_missing_sub_claim_error()
        sentry_sdk.set_user({
            "id": user_id,
            "org_id": claims.get("org_id", None),
            "role": claims.get(CLAIM_ROLE, None),
        })
    except (jwt.InvalidTokenError, RuntimeError):
        # Don't fail authentication if Sentry user setup fails
        # Only catch specific exceptions related to token processing
        pass


def get_token(use_cache: bool = True, use_device_flow: bool = False) -> str:
    """Retrieves an authentication token, either from cache or via login.

    Args:
        use_cache (bool): Boolean indicating whether to store & use the token from disk cache.
            Defaults to True.
        use_device_flow (bool): Boolean indicating whether to force the usage of the device flow for authentication.

    Returns:
        str: The JWT access token.

    Raises:
        RuntimeError: If token retrieval fails.
    """
    token = None

    # Try to get token from cache first
    if use_cache and settings().token_file.exists():
        stored_token = Path(settings().token_file).read_text(encoding="utf-8")
        # Parse stored string "token:expiry_timestamp"
        parts = stored_token.split(":")
        cached_token, expiry_str = parts
        expiry = datetime.fromtimestamp(int(expiry_str), tz=UTC)

        # Check if token is still valid (with some buffer time)
        if datetime.now(tz=UTC) + timedelta(minutes=5) < expiry:
            token = cached_token

    # If no valid cached token, authenticate to get a new one
    if token is None:
        # If we end up here, we:
        # 1. Do not want to use the cached token
        # 2. The cached token is expired
        # 3. No token was cached yet
        token = _authenticate(use_device_flow)
        claims = verify_and_decode_token(token)

        # Store new token with expiry
        if use_cache:
            timestamp = claims["exp"]
            settings().token_file.parent.mkdir(parents=True, exist_ok=True)
            Path(settings().token_file).write_text(f"{token}:{timestamp}", encoding="utf-8")

    # Inform Sentry about the authenticated user (regardless of token source)
    _inform_sentry_about_user(token)

    return token


def remove_cached_token() -> bool:
    """Removes the cached authentication token.

    Deletes the token file if it exists, effectively logging out the user.

    Returns:
        bool: True if the token file was successfully removed, False if it did not exist.
    """
    if settings().token_file.exists():
        settings().token_file.unlink(missing_ok=True)
        return True
    return False


def _authenticate(use_device_flow: bool) -> str:
    """Allows the user to authenticate and obtain an access token.

    Determines the appropriate authentication flow based on whether
    a browser can be opened, then executes that flow.

    Args:
        use_device_flow (bool): Boolean indicating whether to force the usage of the device flow for authentication.

    Returns:
        str: The JWT access token.

    Raises:
        RuntimeError: If authentication fails.
        AssertionError: If the returned token doesn't have the expected format.
    """
    if refresh_token := settings().refresh_token:
        token = _token_from_refresh_token(refresh_token)
    elif _can_open_browser() and not use_device_flow:
        token = _perform_authorization_code_with_pkce_flow()
    else:
        token = _perform_device_flow()
    if not token:
        raise RuntimeError(AUTHENTICATION_FAILED)
    return token


def verify_and_decode_token(token: str) -> dict[str, str]:
    """
    Verifies and decodes the JWT token using the public key from JWS JSON URL.

    Args:
        token (str): The JWT token to verify and decode.

    Returns:
        dict[str,str]: The decoded token claims.

    Raises:
        RuntimeError: If token verification or decoding fails.
    """
    jwk_client = jwt.PyJWKClient(settings().jws_json_url)
    try:
        # Get the public key from the JWK client
        key = jwk_client.get_signing_key_from_jwt(token).key

        # Verify and decode the token using the public key
        decoded = t.cast(
            "dict[str, str]",
            jwt.decode(token, key=key, algorithms=["RS256"], audience=settings().audience),
        )
        if "iss" in decoded and decoded["iss"] != settings().issuer:
            message = f"Token issuer mismatch: expected {settings().issuer}, got {decoded['iss']}"
            raise RuntimeError(message)
        return decoded
    except jwt.exceptions.PyJWTError as e:
        raise RuntimeError(AUTHENTICATION_FAILED) from e


def userinfo(token: str) -> dict[str, str]:
    """Fetches user information from the Auth0 API using the provided token.

    Args:
        token (str): The JWT access token.

    Returns:
        dict[str, str]: The user information.

    Raises:
        RuntimeError: If fetching user information fails.
    """
    try:
        return Users("aignostics-platform.eu.auth0.com").userinfo(access_token=token)  # type: ignore[no-any-return]
    except Exception as e:
        message = f"Failed to fetch user information: {e!s}"
        raise RuntimeError(message) from e


def _can_open_browser() -> bool:
    """Checks if a browser can be opened for authentication.

    Returns:
        bool: True if a browser can be opened, False otherwise.
    """
    launch_browser = False
    try:
        _ = webbrowser.get()
        launch_browser = True
    except webbrowser.Error:
        launch_browser = False

    return launch_browser


def _perform_authorization_code_with_pkce_flow() -> str:
    """Performs the OAuth 2.0 Authorization Code flow with PKCE.

    Opens a browser for user authentication and uses a local redirect
    to receive the authorization code.

    Returns:
        str: The JWT access token.

    Raises:
        RuntimeError: If authentication fails.
    """
    session = OAuth2Session(
        settings().client_id_interactive.get_secret_value(),
        scope=settings().scope_elements,
        redirect_uri=settings().redirect_uri,
        pkce="S256",
    )
    authorization_url, _ = session.authorization_url(
        settings().authorization_base_url,
        access_type="offline",
        audience=settings().audience,
    )

    authentication_result = AuthenticationResult()

    class OAuthCallbackHandler(BaseHTTPRequestHandler):
        def do_GET(self) -> None:  # noqa: N802
            parsed = parse.urlparse(self.path)
            query = parse.parse_qs(parsed.query)

            if "code" not in query:
                self.send_response(400)
                self.send_header("Content-type", "text/html")
                self.end_headers()
                self.wfile.write(b"Error: No authorization code received")
                authentication_result.error = "No authorization code received"
                return

            auth_code = query["code"][0]
            try:
                # Exchange code for token
                token = session.fetch_token(settings().token_url, code=auth_code, include_client_id=True)
                # Store the token
                authentication_result.token = token["access_token"]
                # Send success response
                self.send_response(200)
                self.send_header("Content-type", "text/html")
                self.end_headers()
                self.wfile.write(b"""
                    <html>
                        <script type="application/javascript">
                            setTimeout(function() { window.close(); }, 1000);
                        </script>
                        <body>
                            <h1>Authentication Successful!</h1>
                            <p>You can close this window now.</p>
                        </body>
                    </html>
                """)

            # we want to catch all exceptions here, so we can display them in the browser
            except Exception as e:  # noqa: BLE001
                # Display error message in browser
                self.send_response(500)
                self.send_header("Content-type", "text/html")
                self.end_headers()
                self.wfile.write(f"Error: {e!s}".encode())
                authentication_result.error = str(e)

        # Silence server logs
        def log_message(self, _format: str, *_args) -> None:  # type: ignore[no-untyped-def] # noqa: PLR6301
            return

    # Create and start the server
    parsed_redirect = parse.urlparse(settings().redirect_uri)
    host, port = parsed_redirect.hostname, parsed_redirect.port
    if not host or not port:
        raise RuntimeError(INVALID_REDIRECT_URI)
    # check if port is callback port is available
    port_unavailable_msg = f"Port {port} is already in use. Free the port, or use the device flow."
    if not _ensure_local_port_is_available(port):
        raise RuntimeError(port_unavailable_msg)
    # start the server
    try:
        with HTTPServer((host, port), OAuthCallbackHandler) as server:
            # Call Auth0 with challenge and redirect to localhost with code after successful authN
            webbrowser.open_new(authorization_url)
            # Extract authorization_code from redirected request, see: OAuthCallbackHandler
            server.handle_request()
    except OSError as e:
        if e.errno == errno.EADDRINUSE:
            raise RuntimeError(port_unavailable_msg) from e
        raise RuntimeError(AUTHENTICATION_FAILED) from e

    if authentication_result.error or not authentication_result.token:
        raise RuntimeError(AUTHENTICATION_FAILED)

    return authentication_result.token


def _perform_device_flow() -> str | None:
    """Performs the OAuth 2.0 Device Authorization flow.

    Used when a browser cannot be opened. Provides a URL for the user to visit
    on another device and polls for authorization completion.

    Returns:
        str | None: The JWT access token.

    Raises:
        RuntimeError: If authentication fails or is denied.
    """
    response = requests.post(
        settings().device_url,
        data={
            "client_id": settings().client_id_device.get_secret_value(),
            "scope": settings().scope_elements,
            "audience": settings().audience,
        },
        timeout=settings().request_timeout_seconds,
    )
    try:
        response.raise_for_status()
        json_response = response.json()
        device_code = json_response["device_code"]
        verification_uri = json_response["verification_uri_complete"]
        user_code = json_response["user_code"]
        interval = int(json_response["interval"])
        print(
            f"Your user code is: {user_code}.\nPlease visit: {verification_uri} and verify the same code is displayed!"
        )

    except HTTPError as e:
        raise RuntimeError(AUTHENTICATION_FAILED) from e

    # Polling for access token with received device code
    while True:
        try:
            json_response = requests.post(
                settings().token_url,
                headers={"Accept": "application/json"},
                data={
                    "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
                    "device_code": device_code,
                    "client_id": settings().client_id_device.get_secret_value(),
                },
                timeout=settings().request_timeout_seconds,
            ).json()
            if "error" in json_response:
                if json_response["error"] in {"authorization_pending", "slow_down"}:
                    time.sleep(interval)
                    continue
                raise RuntimeError(AUTHENTICATION_FAILED)

            return t.cast("str", json_response["access_token"])
        except requests.exceptions.JSONDecodeError as e:
            # Handle case where response is not JSON
            raise RuntimeError(AUTHENTICATION_FAILED) from e
        except HTTPError as e:
            raise RuntimeError(AUTHENTICATION_FAILED) from e


def _token_from_refresh_token(refresh_token: SecretStr) -> str | None:
    """Obtains a new access token using a refresh token.

    Args:
        refresh_token (SecretStr): The refresh token to use for obtaining a new access token.

    Returns:
        str | None: The new JWT access token.

    Raises:
        RuntimeError: If token refresh fails.
    """
    try:
        response = requests.post(
            settings().token_url,
            headers={"Accept": "application/json"},
            data={
                "grant_type": "refresh_token",
                "client_id": settings().client_id_interactive.get_secret_value(),
                "refresh_token": refresh_token.get_secret_value(),
            },
            timeout=settings().request_timeout_seconds,
        )
        response.raise_for_status()
        return t.cast("str", response.json()["access_token"])
    except (HTTPError, requests.exceptions.RequestException) as e:
        raise RuntimeError(AUTHENTICATION_FAILED) from e


def _ensure_local_port_is_available(port: int, max_retries: int = CALLBACK_PORT_RETRY_COUNT) -> bool:
    """Check if a port is already in use.

    Args:
        port (int): The port number to check.
        max_retries (int): The maximum number of retries to check the port.

    Returns:
        bool: True if the port is not in use, False otherwise.
    """

    def is_port_available() -> bool:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            try:
                s.bind(("localhost", port))
                return True
            except OSError:
                return False

    retry_count = 0
    while not is_port_available() and retry_count < max_retries:
        time.sleep(1)
        retry_count += 1
    return retry_count < max_retries



"""GUI of dataset module."""

from multiprocessing import Manager
from pathlib import Path

from showinfm.showinfm import show_in_file_manager

from aignostics.gui import frame
from aignostics.utils import get_user_data_directory

from ..utils import BasePageBuilder, GUILocalFilePicker  # noqa: TID252
from ._service import TARGET_LAYOUT_DEFAULT, Service

MESSAGE_NO_DOWNLOAD_FOLDER_SELECTED = "No download folder selected"
PORTAL_URL = "https://portal.imaging.datacommons.cancer.gov/explore/"
SOURCE_EXAMPLE_ID = "1.3.6.1.4.1.5962.99.1.1069745200.1645485340.1637452317744.2.0"


class PageBuilder(BasePageBuilder):
    @staticmethod
    def register_pages() -> None:  # noqa: C901, PLR0915 #NOSONAR
        import nicegui  # noqa: PLC0415
        from nicegui import app, binding, run, ui  # noqa: PLC0415
        from nicegui.events import ValueChangeEventArguments  # noqa: PLC0415

        app.add_static_files("/dataset_assets", Path(__file__).parent / "assets")

        download_message_queue = None

        @binding.bindable_dataclass
        class DownloadForm:
            """Download."""

            source: str | None = None
            destination: Path | None = None
            download_button: ui.button | None = None
            download_arrow_icon: ui.icon | None = None
            download_progress: ui.circular_progress | None = None
            destination_label: ui.label | None = None
            destination_open_button: ui.button | None = None

        download_form = DownloadForm()

        @ui.page("/dataset/idc")
        async def page_idc() -> None:  # noqa: C901, PLR0915, RUF029
            """IDC page."""
            with frame("Download Datasets from Image Data", left_sidebar=False):
                # No need to do anything here
                pass
            with ui.row(align_items="start").classes("full-width"):
                ui.markdown(
                    """
                    ### Download public DICOM datasets
                    Explore the IDC Portal from NCI to find datasets of interest and upload them to the Launchpad
                    to test Aignostics applications.
                    """
                ).classes("w-3/5")
                ui.space()
                with ui.column().classes("w-1/5"):
                    ui.image("/dataset_assets/NIH-IDC-logo.svg").classes("w-25").style("margin-top:1.25rem")
                    with ui.link(target=PORTAL_URL, new_tab=True), ui.button("Explore Portal", icon="search"):
                        ui.tooltip("Explore the IDC Portal to find datasets of interest")

            with ui.row(align_items="start").classes("w-full"):
                ui.markdown("""
                    ##### Pick a Dataset
                    1. Click “Explore Portal” on the right and find a DICOM dataset of interest
                    2. Copy the UID into the field below
                    """).classes("w-2/5")
                ui.space()
                ui.label("OR").style("margin-top: 2rem; font-size: 2rem; color: grey")
                ui.space()
                ui.markdown("""
                    ##### Quick Start
                    1. Click “Use Example Dataset” below to automatically populate a pre-selected DICOM dataset
                    """).classes("w-2/5")

            with ui.row(align_items="start").classes("w-full"):
                ui.markdown(
                    """
                    Once your dataset UID is added, click "Data" or Custom and select a
                    folder on your computer for the dataset to be stored.

                    Once you've chosen, the “Download” button should light up and you can download your
                    selected dataset.

                    When you're ready to begin image analysis, open ☰ menu and click “Run Applications” to return to
                    the main menu.
                    """
                )

            def _on_source_input_change(e: ValueChangeEventArguments) -> None:
                """On change event."""
                if download_form.download_button is None:
                    return
                if e.value:
                    download_form.source = e.value
                else:
                    download_form.source = None
                if (download_form.source is not None) and (download_form.destination is not None):
                    download_form.download_button.enable()
                else:
                    download_form.download_button.disable()

            async def _select_destination() -> None:
                """Open a file picker dialog and show notifier when closed again."""
                if (
                    download_form.destination_label is None
                    or download_form.destination_open_button is None
                    or download_form.download_button is None
                ):
                    return

                result = await GUILocalFilePicker(str(Path.home()), multiple=False)  # type: ignore
                if result and len(result) > 0:
                    path = Path(result[0])
                    if not path.is_dir():
                        download_form.destination = None
                        download_form.destination_label.set_text(MESSAGE_NO_DOWNLOAD_FOLDER_SELECTED)
                        download_form.destination_open_button.disable()
                        ui.notify(
                            "The selected path is not a directory. Please select a valid directory.", type="warning"
                        )
                    else:
                        download_form.destination = path
                        download_form.destination_label.set_text(str(path))
                        download_form.destination_open_button.enable()
                        ui.notify(f"You chose directory {download_form.destination}.", type="info")
                else:
                    download_form.destination = None
                    download_form.destination_label.set_text(MESSAGE_NO_DOWNLOAD_FOLDER_SELECTED)
                    download_form.destination_open_button.disable()
                    ui.notify("You did not make a selection. You must choose a download folder.", type="warning")
                if (download_form.source is not None) and (download_form.destination is not None):
                    download_form.download_button.enable()
                else:
                    download_form.download_button.disable()

            async def _select_data() -> None:  # noqa: RUF029
                """Open a file picker dialog and show notifier when closed again."""
                if (
                    download_form.destination_label is None
                    or download_form.destination_open_button is None
                    or download_form.download_button is None
                ):
                    return

                download_form.destination = get_user_data_directory("datasets/idc")
                download_form.destination_label.set_text(str(download_form.destination))
                download_form.destination_open_button.enable()
                if download_form.source is not None:
                    download_form.download_button.enable()
                else:
                    download_form.download_button.disable()

            def _open_destination() -> None:
                """Open the destination directory in the file explorer."""
                show_in_file_manager(str(download_form.destination))

            async def _download(source: str) -> None:
                """Download."""
                if (
                    download_form.destination is None
                    or download_form.download_button is None
                    or download_form.download_progress is None
                    or download_form.destination is None
                    or download_form.download_arrow_icon is None
                ):
                    return
                nonlocal download_message_queue
                ui.notify(f"Downloading {source!s} ...", type="info")
                download_form.download_arrow_icon.visible = False
                download_form.download_progress.visible = True
                download_form.download_button.props(add="loading")
                if download_message_queue is None:
                    download_message_queue = Manager().Queue()
                try:
                    await run.io_bound(
                        Service.download_with_queue,  # type: ignore[arg-type]
                        download_message_queue,  # type: ignore[unused-ignore] # type: ignore
                        source,
                        str(download_form.destination),
                        TARGET_LAYOUT_DEFAULT,
                        False,
                    )
                except Exception as e:  # noqa: BLE001
                    nicegui.ui.notify(f"Download failed: {e}", type="negative", multi_line=True)
                    download_form.download_button.props(remove="loading")
                    download_form.download_progress.visible = False
                    download_form.download_arrow_icon.visible = False
                    return
                ui.notify("Download completed.", type="positive")
                download_form.download_button.props(remove="loading")
                download_form.download_progress.visible = False
                download_form.download_arrow_icon.visible = True
                _open_destination()

            with ui.card().classes("w-full"):
                ui.label("Download Dataset").classes("text-h6")
                with ui.row(align_items="center").classes("w-full"):
                    source_input = (
                        ui.input(
                            label="Enter ID of collection, patient case, study, series or instance.",
                            placeholder="Click 🔍 Explore Portal to find IDs",
                            on_change=lambda e: _on_source_input_change(e),  # noqa: PLW0108
                        )
                        .props("clearable")
                        .classes("w-2/5")
                        .mark("SOURCE_INPUT")
                    )
                    ui.space()
                    download_form.download_arrow_icon = ui.icon(name="east", size="lg", color="primary")
                    download_form.download_progress = ui.circular_progress(show_value=False).props("instant-feedback")
                    with download_form.download_progress:
                        ui.button(icon="cloud_download").props("flat round").disable()
                    download_form.download_progress.visible = False
                    ui.space()
                    with ui.row(align_items="center").classes("w-2/5"):
                        ui.space()
                        download_form.destination_label = ui.label(
                            MESSAGE_NO_DOWNLOAD_FOLDER_SELECTED
                            if download_form.destination is None
                            else str(download_form.destination)
                        ).classes("max-w-72")
                        download_form.destination_open_button = ui.button(
                            icon="folder_open", on_click=_open_destination, color="secondary"
                        )
                        download_form.destination_open_button.mark("BUTTON_OPEN_DESTINATION").disable()

                with ui.row(align_items="center").classes("w-full"):
                    with ui.button(
                        "Example Dataset",
                        on_click=lambda _: source_input.set_value(SOURCE_EXAMPLE_ID),
                        icon="folder",
                        color="secondary",
                    ).mark("BUTTON_EXAMPLE_DATASET"):
                        ui.tooltip("Use a pre-selected example dataset")
                    ui.space()
                    with ui.button("Data", on_click=_select_data, icon="folder_special", color="purple-400").mark(
                        "BUTTON_DOWNLOAD_DESTINATION_DATA"
                    ):
                        ui.tooltip("Use Launchpad datasets directory")

                    with ui.button("Custom", on_click=_select_destination, icon="folder").mark(
                        "BUTTON_DOWNLOAD_DESTINATION"
                    ):
                        ui.tooltip("Select a custom directory")
                    ui.space()
                    with ui.row(align_items="center"):
                        with ui.button("Download", icon="cloud_download").mark("BUTTON_DOWNLOAD") as download_button:
                            ui.tooltip("Download the selected dataset")
                        download_form.download_button = download_button
                        download_form.download_button.on("click", lambda _: _download(source_input.value))
                        download_form.download_button.disable()

            def update_progress() -> None:
                """Update the progress indicator with values from the queue."""
                if download_form.download_progress is None:
                    return

                while download_message_queue and not download_message_queue.empty():
                    new_value = download_message_queue.get()
                    download_form.download_progress.set_value(new_value)

            ui.timer(0.1, update_progress)



"""Example script to run the test application with one example image."""

import tempfile

from aignostics import platform

# initialize the client
client = platform.Client()
# create application run
# for details, see the IPython or Marimo notebooks for a detailed explanation of the payload
application_run = client.runs.create(
    application_version="two-task-dummy:v0.35.0",
    items=[
        platform.InputItem(
            reference="1",
            input_artifacts=[
                platform.InputArtifact(
                    name="user_slide",
                    download_url=platform.generate_signed_url(
                        "gs://aignx-storage-service-dev/sample_data_formatted/9375e3ed-28d2-4cf3-9fb9-8df9d11a6627.tiff"
                    ),
                    metadata={
                        "checksum_crc32c": "N+LWCg==",
                        "base_mpp": 0.46499982,
                        "width": 3728,
                        "height": 3640,
                    },
                )
            ],
        ),
    ],
)
# wait for the results and download incrementally as they become available
tmp_folder = tempfile.gettempdir()
application_run.download_to_folder(tmp_folder)



"""Static configuration of Aignostics Python SDK."""

import os
from pathlib import Path

# Configuration required by oe-python-template
API_VERSIONS: dict[str, str] = {"v1": "1.0.0"}
MODULES_TO_INSTRUMENT: list[str] = ["aignostics.qupath"]
NOTEBOOK_DEFAULT = Path(__file__).parent / "notebook" / "_notebook.py"

# Project specific configuration
os.environ["MATPLOTLIB"] = "false"
os.environ["NICEGUI_STORAGE_PATH"] = str(Path.home().resolve() / ".aignostics" / ".nicegui")
WSI_SUPPORTED_FILE_EXTENSIONS = {".dcm", ".tiff", ".tif", ".svs"}



"""CLI of dataset module."""

import webbrowser
from pathlib import Path
from typing import Annotated

import requests
import typer

from aignostics.platform import generate_signed_url as platform_generate_signed_url
from aignostics.utils import console, get_logger, get_user_data_directory

logger = get_logger(__name__)

PATH_LENFTH_MAX = 260
TARGET_LAYOUT_DEFAULT = "%collection_id/%PatientID/%StudyInstanceUID/%Modality_%SeriesInstanceUID/"

cli = typer.Typer(
    name="dataset",
    help="Download datasets from National Institute of Cancer (NIC) and Aignostics.",
)

idc_app = typer.Typer()
cli.add_typer(
    idc_app,
    name="idc",
    help="Download public datasets from Image Data Commons (IDC) Portal of National Institute of Cancer (NIC).",
)

aignostics_app = typer.Typer()
cli.add_typer(aignostics_app, name="aignostics", help="Download proprietary sample datasets from Aignostics.")


@idc_app.command()
def browse() -> None:
    """Open browser to explore IDC portal."""
    webbrowser.open("https://portal.imaging.datacommons.cancer.gov/explore/")


@idc_app.command()
def indices() -> None:
    """List available columns in given of the IDC Portal."""
    from aignostics.third_party.idc_index import IDCClient  # noqa: PLC0415

    client = IDCClient.client()
    console.print(list(client.indices_overview.keys()))


@idc_app.command()
def columns(
    index: Annotated[
        str,
        typer.Option(
            help="List available columns in given of the IDC Portal."
            " See List available columns in given of the IDC Portal for available indices"
        ),
    ] = "sm_instance_index",
) -> None:
    """List available columns in given of the IDC Portal."""
    from aignostics.third_party.idc_index import IDCClient  # noqa: PLC0415

    client = IDCClient.client()
    client.fetch_index(index)
    console.print(list(getattr(client, index).columns))


@idc_app.command()
def query(
    query: Annotated[
        str,
        typer.Argument(
            help="SQL Query to execute."
            "See https://idc-index.readthedocs.io/en/latest/column_descriptions.html "
            "for indices and their attributes"
        ),
    ] = """SELECT
    SOPInstanceUID, SeriesInstanceUID, ImageType[3], instance_size, TotalPixelMatrixColumns, TotalPixelMatrixRows
FROM
    sm_instance_index
WHERE
    TotalPixelMatrixColumns > 25000
    AND TotalPixelMatrixRows > 25000
    AND ImageType[3] = 'VOLUME'
""",
    indices: Annotated[
        str,
        typer.Option(
            help="Comma separated list of additional indices to sync before running the query."
            " The main index is always present. By default sm_instance_index is synched in addition."
            " See https://idc-index.readthedocs.io/en/latest/column_descriptions.html for available indices."
        ),
    ] = "sm_instance_index",
) -> None:
    """Query IDC index. For example queries see https://github.com/ImagingDataCommons/IDC-Tutorials/blob/master/notebooks/labs/idc_rsna2023.ipynb."""
    import pandas as pd  # noqa: PLC0415

    from aignostics.third_party.idc_index import IDCClient  # noqa: PLC0415

    client = IDCClient.client()
    for idx in [idx.strip() for idx in indices.split(",") if idx.strip()]:
        logger.info("Fetching index: '%s'", idx)
        client.fetch_index(idx)

    pd.set_option("display.max_colwidth", None)
    console.print(client.sql_query(sql_query=query))  # type: ignore[no-untyped-call]


@idc_app.command(name="download")
def idc_download(
    source: Annotated[
        str,
        typer.Argument(
            help="Identifier or comma-separated set of identifiers."
            " IDs matched against collection_id, PatientId, StudyInstanceUID, SeriesInstanceUID or SOPInstanceUID."
        ),
    ],
    target: Annotated[
        Path,
        typer.Argument(
            help="target directory for download",
            exists=True,
            file_okay=False,
            dir_okay=True,
            writable=True,
            readable=True,
            resolve_path=True,
        ),
    ] = get_user_data_directory("datasets/idc"),  # noqa: B008
    target_layout: Annotated[
        str, typer.Option(help="layout of the target directory. See default for available elements for use")
    ] = TARGET_LAYOUT_DEFAULT,
    dry_run: Annotated[bool, typer.Option(help="dry run")] = False,
) -> None:
    """Download from manifest file, identifier, or comma-separate set of identifiers.

    Raises:
        typer.Exit: If the target directory does not exist.
    """
    from aignostics.third_party.idc_index import IDCClient  # noqa: PLC0415

    client = IDCClient.client()
    logger.info("Downloading instance index from IDC version: %s", client.get_idc_version())  # type: ignore[no-untyped-call]

    target_directory = Path(target)
    if not target_directory.is_dir():
        logger.error("Target directory does not exist: %s", target_directory)
        raise typer.Exit(code=1)

    item_ids = [item for item in source.split(",") if item]

    if not item_ids:
        logger.error("No valid IDs provided.")

    index_df = client.index
    client.fetch_index("sm_instance_index")
    logger.info("Downloaded instance index")
    sm_instance_index_df = client.sm_instance_index

    def check_and_download(column_name: str, item_ids: list[str], target_directory: Path, kwarg_name: str) -> bool:
        if column_name != "SOPInstanceUID":
            matches = index_df[column_name].isin(item_ids)
            matched_ids = index_df[column_name][matches].unique().tolist()
        else:
            matches = sm_instance_index_df[column_name].isin(item_ids)  # type: ignore
            matched_ids = sm_instance_index_df[column_name][matches].unique().tolist()  # type: ignore
        if not matched_ids:
            return False
        unmatched_ids = list(set(item_ids) - set(matched_ids))
        if unmatched_ids:
            logger.debug("Partial match for %s: matched %s, unmatched %s", column_name, matched_ids, unmatched_ids)
        logger.info("Identified matching %s: %s", column_name, matched_ids)
        client.download_from_selection(**{  # type: ignore[no-untyped-call]
            kwarg_name: matched_ids,
            "downloadDir": target_directory,
            "dirTemplate": target_layout,
            "quiet": False,
            "show_progress_bar": True,
            "use_s5cmd_sync": True,
            "dry_run": dry_run,
        })
        return True

    matches_found = 0
    matches_found += check_and_download("collection_id", item_ids, target_directory, "collection_id")
    matches_found += check_and_download("PatientID", item_ids, target_directory, "patientId")
    matches_found += check_and_download("StudyInstanceUID", item_ids, target_directory, "studyInstanceUID")
    matches_found += check_and_download("SeriesInstanceUID", item_ids, target_directory, "seriesInstanceUID")
    matches_found += check_and_download("SOPInstanceUID", item_ids, target_directory, "sopInstanceUID")
    if not matches_found:
        logger.error(
            "None of the values passed matched any of the identifiers: "
            "collection_id, PatientID, StudyInstanceUID, SeriesInstanceUID, SOPInstanceUID."
        )


@aignostics_app.command("download")
def aignostics_download(
    source_url: Annotated[
        str,
        typer.Argument(
            help="URL to download, e.g. gs://aignx-storage-service-dev/sample_data_formatted/9375e3ed-28d2-4cf3-9fb9-8df9d11a6627.tiff"
        ),
    ],
    destination_directory: Annotated[
        Path,
        typer.Argument(
            help="Destination directory to download to",
            exists=True,
            file_okay=False,
            dir_okay=True,
            writable=True,
            readable=True,
            resolve_path=True,
        ),
    ] = get_user_data_directory("datasets/aignostics"),  # noqa: B008
) -> None:
    """Download from bucket to folder via a signed URL."""
    from rich.progress import (  # noqa: PLC0415
        BarColumn,
        FileSizeColumn,
        Progress,
        TaskProgressColumn,
        TextColumn,
        TimeRemainingColumn,
        TotalFileSizeColumn,
        TransferSpeedColumn,
    )

    # Get filename from URL
    filename = source_url.split("/")[-1]

    # Generate a signed URL
    source_url_signed = platform_generate_signed_url(source_url)

    output_path = Path(destination_directory) / filename

    console.print(f"Downloading from {source_url} to {output_path}")

    # Make sure the destination directory exists
    Path(destination_directory).mkdir(parents=True, exist_ok=True)

    # Start the request to get content length
    response = requests.get(source_url_signed, stream=True, timeout=60)
    total_size = int(response.headers.get("content-length", 0))

    with Progress(
        TextColumn("[progress.description]Downloading"),
        BarColumn(),
        TaskProgressColumn(),
        TimeRemainingColumn(),
        FileSizeColumn(),
        TotalFileSizeColumn(),
        TransferSpeedColumn(),
        TextColumn("[progress.description]{task.description}"),
    ) as progress:
        # Create a task for overall progress
        task = progress.add_task(f"Downloading {filename}", total=total_size)

        # Write the file
        with open(output_path, "wb") as f:
            for chunk in response.iter_content(chunk_size=8192):
                if chunk:  # filter out keep-alive new chunks
                    f.write(chunk)
                    progress.update(task, advance=len(chunk))

    console.print(f"[green]Successfully downloaded to {output_path}[/green]")



"""Notebook server utilities."""

from collections.abc import Callable
from pathlib import Path
from typing import Any

from ._health import Health
from ._log import get_logger
from ._fs import get_user_data_directory
import shutil

logger = get_logger(__name__)


def register_health_endpoint(router: Any) -> Callable[..., Health]:  # noqa: ANN401
    """Register health endpoint to the given router.

    Args:
        router: The router to register the health endpoint to.

    Returns:
        Callable[..., Health]: The health endpoint function.
    """
    # We accept 'Any' instead of APIRouter to avoid importing fastapi at module level

    @router.get("/healthz")
    def health_endpoint() -> Health:
        """Determine health of the app.

        Returns:
            Health: Health.
        """
        return Health(status=Health.Code.UP)

    # Explicitly type the return value to satisfy mypy
    result: Callable[..., Health] = health_endpoint
    return result


def create_marimo_app(notebook: Path, override_if_exists: bool) -> Any:  # noqa: ANN401
    """Create a FastAPI app with marimo notebook server.

    Args:
        notebook (Path): Path to the notebook. Notebook will be copied into the user data directory.
        override_if_exists (bool): Whether to override the notebook in the user data directory if it already exists.

    Returns:
        FastAPI: FastAPI app with marimo notebook server.

    Raises:
        ValueError: If the notebook directory does not exist.
    """
    # Import dependencies only when function is called
    import marimo  # noqa: PLC0415
    from fastapi import APIRouter, FastAPI  # noqa: PLC0415

    server = marimo.create_asgi_app(include_code=True)
    directory = get_user_data_directory("notebooks")
    notebook_destination = directory / notebook.name
    if notebook_destination.exists() and not override_if_exists:
        logger.debug(f"Notebook already exists at {notebook_destination}, using existing file")
    else:
        notebook_destination.write_bytes(notebook.read_bytes())
        logger.info(f"Copied notebook from {notebook} to {notebook_destination}")
    server = server.with_app(path="/", root=str(notebook_destination.resolve()))
    app = FastAPI()
    router = APIRouter(tags=["marimo"])
    register_health_endpoint(router)
    app.include_router(router)
    app.mount("/", server.build())
    return app



"""Utility functions for application module.

1. Printing of application resources
2. Reading/writing metadata CSV files
3. Mime type handling.
"""

import csv
import mimetypes
from enum import StrEnum
from pathlib import Path
from typing import Any, Literal

from aignostics.platform import (
    ApplicationRun,
    ApplicationRunData,
    ApplicationRunStatus,
    InputArtifactData,
    OutputArtifactData,
    OutputArtifactElement,
)
from aignostics.utils import console, get_logger

logger = get_logger(__name__)

RUN_FAILED_MESSAGE = "Failed to get status for run with ID '%s'"


class OutputFormat(StrEnum):
    """
    Enum representing the supported output formats.

    This enum defines the possible formats for output data:
    - TEXT: Output data as formatted text
    - JSON: Output data in JSON format
    """

    TEXT = "text"
    JSON = "json"


def retrieve_and_print_run_details(run: ApplicationRun) -> None:
    """Retrieve and print detailed information about a run.

    Args:
        run (ApplicationRun): The ApplicationRun object

    """
    run_data = run.details()
    console.print(f"[bold]Run Details for {run.application_run_id}[/bold]")
    console.print("=" * 80)
    console.print(f"[bold]App Version:[/bold] {run_data.application_version_id}")
    console.print(f"[bold]Status:[/bold] {run_data.status.value}")
    console.print(f"[bold]Triggered at:[/bold] {run_data.triggered_at}")
    console.print(f"[bold]Organization:[/bold] {run_data.organization_id}")
    console.print(f"[bold]Triggered by:[/bold] {run_data.triggered_by}")

    # Get and display detailed item status
    console.print()
    console.print("[bold]Items:[/bold]")

    _retrieve_and_print_run_items(run)
    _print_run_status_summary(run)


def _retrieve_and_print_run_items(run: ApplicationRun) -> None:
    """Retrieve and print information about items in a run.

    Args:
        run (ApplicationRun): The ApplicationRun object
    """
    # Get results with detailed information
    results = run.results()
    if not results:
        console.print("  No item results available.")
        return

    for item in results:
        console.print(f"  [bold]Item Reference:[/bold] {item.reference}")
        console.print(f"  [bold]Item ID:[/bold] {item.item_id}")
        console.print(f"  [bold]Status:[/bold] {item.status.value}")

        if item.error:
            console.print(f"  [error]Error:[/error] {item.error}")

        if item.output_artifacts:
            console.print("  [bold]Output Artifacts:[/bold]")
            for artifact in item.output_artifacts:
                console.print(f"    - Name: {artifact.name}")
                console.print(f"      MIME Type: {artifact.mime_type}")
                console.print(f"      Artifact ID: {artifact.output_artifact_id}")

        console.print()


def _print_run_status_summary(run: ApplicationRun) -> None:
    """Print summary of item statuses in a run.

    Args:
        run (ApplicationRun): The ApplicationRun object
    """
    # Get and display item status counts
    item_statuses = run.item_status()
    if not item_statuses:
        return

    status_counts: dict[
        Literal["pending", "canceled_user", "canceled_system", "error_user", "error_system", "succeeded"], int
    ] = {}
    for status in item_statuses.values():
        status_counts[status.value] = status_counts.get(status.value, 0) + 1

    console.print("[bold]Item Status Summary:[/bold]")
    for status, count in status_counts.items():
        console.print(f"  {status}: {count}")


def _retrieve_and_print_item_status_counts(run: ApplicationRun) -> bool:
    """Retrieve and print item status counts for a run.

    Args:
        run (ApplicationRun): The run object

    Returns:
        bool: True if successful, False otherwise
    """
    try:
        item_statuses = run.item_status()
    except Exception as e:
        logger.exception("Failed to get item status for run with ID '%s'", run.application_run_id)
        console.print(
            f"[error]Error:[/error] Failed to get item statuses for run with ID '{run.application_run_id}': {e}"
        )
        return False

    status_counts: dict[
        Literal["pending", "canceled_user", "canceled_system", "error_user", "error_system", "succeeded"], int
    ] = {}
    for status in item_statuses.values():
        status_counts[status.value] = status_counts.get(status.value, 0) + 1

    if status_counts:
        console.print("[bold]Item Status Counts:[/bold]")
        for status, count in status_counts.items():
            console.print(f"  {status}: {count}")

    return True


def print_runs_verbose(runs: list[ApplicationRunData]) -> None:
    """Print detailed information about runs, sorted by triggered_at in descending order.

    Args:
        runs (list[ApplicationRunData]): List of run data
        service (Service): The Service instance to use

    """
    from ._service import Service  # noqa: PLC0415

    console.print("[bold]Application Runs:[/bold]")
    console.print("=" * 80)

    for run in runs:
        console.print(f"[bold]Run ID:[/bold] {run.application_run_id}")
        console.print(f"[bold]App Version:[/bold] {run.application_version_id}")
        console.print(f"[bold]Status:[/bold] {run.status.value}")
        console.print(f"[bold]Triggered at:[/bold] {run.triggered_at.astimezone().strftime('%Y-%m-%d %H:%M:%S %Z')}")
        console.print(f"[bold]Organization:[/bold] {run.organization_id}")

        try:
            _retrieve_and_print_item_status_counts(Service().application_run(run.application_run_id))
        except Exception as e:
            logger.exception("Failed to retrieve item status counts for run with ID '%s'", run.application_run_id)
            console.print(
                f"[error]Error:[/error] Failed to retrieve item status counts for run with ID "
                f"'{run.application_run_id}': {e}"
            )
            continue
        console.print("-" * 80)


def print_runs_non_verbose(runs: list[ApplicationRunData]) -> None:
    """Print simplified information about runs, sorted by triggered_at in descending order.

    Args:
        runs (list[ApplicationRunData]): List of runs

    """
    console.print("[bold]Application Run IDs:[/bold]")

    for run_status in runs:
        console.print(
            f"- [bold]{run_status.application_run_id}[/bold] of "
            f"[bold]{run_status.application_version_id}[/bold] "
            f"(triggered: {run_status.triggered_at.astimezone().strftime('%Y-%m-%d %H:%M:%S %Z')}, "
            f"status: {run_status.status.value})"
        )


def write_metadata_dict_to_csv(
    metadata_csv: Path,
    metadata_dict: list[dict[str, Any]],
) -> Path:
    """Write metadata dict to a CSV file.

    Convert dict to CSV including header assuming all entries in dict have the same keys

    Args:
        metadata_csv (Path): Path to the CSV file
        metadata_dict (list[dict[str,Any]]): List of dictionaries containing metadata

    Returns:
        Path: Path to the CSV file
    """
    with metadata_csv.open("w", newline="", encoding="utf-8") as f:
        field_names = list(metadata_dict[0].keys())
        writer = csv.writer(f, delimiter=";", quotechar='"', quoting=csv.QUOTE_MINIMAL)
        writer.writerow(field_names)
        for entry in metadata_dict:
            writer.writerow([entry.get(field, "") for field in field_names])
    return metadata_csv


def read_metadata_csv_to_dict(
    metadata_csv_file: Path,
) -> list[dict[str, Any]] | None:
    """Read metadata CSV file and convert it to a list of dictionaries.

    Args:
        metadata_csv_file (Path): Path to the CSV file

    Returns:
        list[dict[str, str]] | None: List of dictionaries containing metadata or None if an error occurs
    """
    try:
        with metadata_csv_file.open("r", encoding="utf-8") as f:
            return list(csv.DictReader(f, delimiter=";", quotechar='"'))
    except (csv.Error, UnicodeDecodeError, KeyError) as e:
        logger.warning("Failed to parse metadata CSV file '%s': %s", metadata_csv_file, e)
        console.print(f"[warning]Warning:[/warning] Failed to parse metadata CSV file '{metadata_csv_file}': {e}")
        return None


def application_run_status_to_str(
    status: ApplicationRunStatus,
) -> str:
    """Convert application status to a human-readable string.

    Args:
        status (ApplicationRunStatus): The application status

    Raises:
        RuntimeError: If the status is invalid or unknown

    Returns:
        str: Human-readable string representation of the status
    """
    status_mapping = {
        ApplicationRunStatus.CANCELED_SYSTEM: "canceled by platform",
        ApplicationRunStatus.CANCELED_USER: "canceled by user",
        ApplicationRunStatus.COMPLETED: "completed",
        ApplicationRunStatus.COMPLETED_WITH_ERROR: "completed with error",
        ApplicationRunStatus.RECEIVED: "received by platform",
        ApplicationRunStatus.REJECTED: "rejected by platform",
        ApplicationRunStatus.RUNNING: "running on platform",
        ApplicationRunStatus.SCHEDULED: "scheduled for processing",
    }

    if status in status_mapping:
        return status_mapping[status]

    message = f"Unknown application status: {status.value}"
    logger.error(message)
    raise RuntimeError(message)


def get_mime_type_for_artifact(artifact: OutputArtifactData | InputArtifactData | OutputArtifactElement) -> str:
    """Get the MIME type for a given artifact.

    Args:
        artifact (OutputArtifact | InputArtifact | OutputArtifactElement): The artifact to get the MIME type for.

    Returns:
        str: The MIME type of the artifact.
    """
    if isinstance(artifact, InputArtifactData):
        return str(artifact.mime_type)
    if isinstance(artifact, OutputArtifactData):
        return str(artifact.mime_type)
    metadata = artifact.metadata or {}
    return str(metadata.get("media_type", metadata.get("mime_type", "application/octet-stream")))


def get_file_extension_for_artifact(artifact: OutputArtifactData) -> str:
    """Get the file extension for a given artifact.

    Returns .bin if no known extension is found for mime type.

    Args:
        artifact (OutputArtifact): The artifact to get the extension for.

    Returns:
        str: The file extension of the artifact.
    """
    mimetypes.init()
    mimetypes.add_type("application/vnd.apache.parquet", ".parquet")
    mimetypes.add_type("application/geo+json", ".json")

    file_extension = mimetypes.guess_extension(get_mime_type_for_artifact(artifact))
    if file_extension == ".geojson":
        file_extension = ".json"
    if not file_extension:
        file_extension = ".bin"
    logger.debug("Guessed file extension: '%s' for artifact '%s'", file_extension, artifact.name)
    return file_extension



"""Service of the bucket module."""

from collections.abc import Callable, Generator
from pathlib import Path
from typing import Any, cast

import requests
from botocore.client import BaseClient, Config
from botocore.exceptions import ClientError

from aignostics.utils import UNHIDE_SENSITIVE_INFO, BaseService, Health, get_logger

from ._settings import Settings

logger = get_logger(__name__)

BUCKET_PROTOCOL = "gs"
SIGNATURE_VERSION = "s3v4"
ENDPOINT_URL_DEFAULT = "https://storage.googleapis.com"


class Service(BaseService):
    """Service of the bucket module."""

    _settings: Settings

    def __init__(self) -> None:
        """Initialize service."""
        super().__init__(Settings)

    def info(self, mask_secrets: bool = True) -> dict[str, Any]:
        """Determine info of this service.

        Args:
            mask_secrets (bool): If True, mask sensitive information in the output.

        Returns:
            dict[str,Any]: The info of this service.
        """
        return {"settings": self._settings.model_dump(context={UNHIDE_SENSITIVE_INFO: not mask_secrets})}

    def health(self) -> Health:  # noqa: PLR6301
        """Determine health of this service.

        Returns:
            Health: The health of the service.
        """
        return Health(
            status=Health.Code.UP,
            components={},
        )

    def _get_s3_client(self, endpoint_url: str = ENDPOINT_URL_DEFAULT) -> BaseClient:
        """Get a Boto3 S3 client instance for cloud bucket on Aignostics Platform.

        Args:
            endpoint_url (str): The endpoint URL for the S3 service.

        Returns:
            BaseClient: A Boto3 S3 client instance.
        """
        from boto3 import Session  # noqa: PLC0415

        # https://www.kmp.tw/post/accessgcsusepythonboto3/
        session = Session(
            aws_access_key_id=self._settings.hmac_access_key_id.get_secret_value(),
            aws_secret_access_key=self._settings.hmac_secret_access_key.get_secret_value(),
            region_name=self._settings.region_name,
        )
        return session.client("s3", endpoint_url=endpoint_url, config=Config(signature_version=SIGNATURE_VERSION))

    @staticmethod
    def get_bucket_protocol() -> str:
        """Get the bucket protocol.

        Returns:
            str: The bucket protocol.
        """
        return BUCKET_PROTOCOL

    def get_bucket_name(self) -> str:
        """Get the bucket name.

        Returns:
            str: The bucket name.
        """
        return self._settings.name

    def create_signed_upload_url(self, object_key: str, bucket_name: str | None = None) -> str:
        """Generates a signed URL to upload a Google Cloud Storage object.

        Args:
            object_key (str): The key of the object to generate a signed URL for.
            bucket_name (str): The name of the bucket to generate a signed URL for.
                If None, use the default bucket.

        Returns:
            str: A signed URL that can be used to upload to the bucket and key.
        """
        url = self._get_s3_client().generate_presigned_url(
            ClientMethod="put_object",
            Params={"Bucket": self._settings.name if bucket_name is None else bucket_name, "Key": object_key},
            ExpiresIn=self._settings.upload_signed_url_expiration_seconds,
        )
        return cast("str", url)

    def upload_file(
        self, source_path: Path, object_key: str, callback: Callable[[int, Path], None] | None = None
    ) -> bool:
        """Upload a file to the bucket using a signed URL.

        Args:
            source_path (Path): Path of the local file to upload.
            object_key (str): Key to use for the uploaded object.
            callback (Callable[[int, int], None] | None): Optional callback function for upload progress.
                Function receives bytes_read and total_bytes parameters.

        Returns:
            bool: True if upload was successful, False otherwise.
        """
        logger.debug("Uploading file '%s' to object key '%s'", source_path, object_key)
        if not source_path.is_file():
            logger.error("Source path '%s' is not a file", source_path)
            return False

        signed_url = self.create_signed_upload_url(object_key)

        try:
            with open(source_path, "rb") as f:

                def read_in_chunks() -> Generator[bytes, None, None]:
                    while True:
                        chunk = f.read(1048576)  # ~1MB chunks
                        if not chunk:
                            break
                        if callback:
                            callback(len(chunk), source_path)
                        yield chunk

                response = requests.put(
                    signed_url,
                    data=read_in_chunks(),
                    headers={"Content-Type": "application/octet-stream"},
                    timeout=60,
                )
                response.raise_for_status()

            logger.info("Successfully uploaded '%s' to object key '%s'", source_path, object_key)
            return True

        except (OSError, requests.RequestException):
            logger.exception("Error uploading file '%s' to object key '%s'", source_path, object_key)
            return False

    def create_signed_download_url(self, object_key: str, bucket_name: str | None = None) -> str:
        """Generates a signed URL to download a Google Cloud Storage object.

        Args:
            object_key (str): The key of the object to generate a signed URL for.
            bucket_name (str | None): The name of the bucket to generate a signed URL for.
                If None, use the default bucket.

        Returns:
            str: A signed URL that can be used to download from the bucket and key.
        """
        url = self._get_s3_client().generate_presigned_url(
            ClientMethod="get_object",
            Params={"Bucket": self._settings.name if bucket_name is None else bucket_name, "Key": object_key},
            ExpiresIn=self._settings.download_signed_url_expiration_seconds,
        )
        return cast("str", url)

    def upload(
        self,
        source_path: Path,
        destination_prefix: str,
        callback: Callable[[int, Path], None] | None = None,
    ) -> dict[str, list[str]]:
        """Upload a file or directory to the bucket.

        Args:
            source_path (Path): Path to file or directory to upload.
            destination_prefix (str): Prefix for object keys (e.g. username).
            callback (Callable[[int, int], None] | None): Optional callback function for upload progress.
                Function receives bytes_read and total_bytes parameters.

        Returns:
            dict[str, list[str]]: Dict with 'success' and 'failed' lists containing object keys.
        """
        results: dict[str, list[str]] = {"success": [], "failed": []}

        destination_prefix = destination_prefix.rstrip("/")

        if source_path.is_file():
            object_key = f"{destination_prefix}/{source_path.name}"
            if self.upload_file(source_path, object_key, callback):
                results["success"].append(object_key)
            else:
                results["failed"].append(object_key)

        elif source_path.is_dir():
            for file_path in source_path.glob("**/*"):
                if file_path.is_file():
                    rel_path = file_path.relative_to(source_path).as_posix()
                    object_key = f"{destination_prefix}/{rel_path}"

                    if self.upload_file(file_path, object_key, callback):
                        results["success"].append(object_key)
                    else:
                        results["failed"].append(object_key)
        else:
            logger.error("Source path '%s' is neither a file nor directory", source_path)

        return results

    def ls(self, detail: bool = False) -> list[str | dict[str, Any]]:
        """List objects directly in the bucket (non-recursive).

        Args:
            detail (bool): If True, return detailed information including object type, else return only paths.

        Returns:
            list[Union[str, dict[str, Any]]]: List of objects directly in the bucket with optional detail.
        """
        s3c = self._get_s3_client()
        response = s3c.list_objects_v2(Bucket=self._settings.name, Prefix="", Delimiter="/")

        result: list[str | dict[str, Any]] = []
        contents = response.get("Contents", [])
        common_prefixes = response.get("CommonPrefixes", [])

        if detail:
            # Process directories (common prefixes)
            for prefix in common_prefixes:
                if prefix.get("Prefix") not in {None, ""}:
                    prefix_path = f"{self._settings.name}/{prefix['Prefix']}"
                    result.append({
                        "key": prefix_path,
                        "size": 0,
                        "last_modified": None,
                        "etag": "",
                        "storage_class": "",
                        "type": "directory",
                    })

            # Process files
            for item in contents:
                if item.get("Key") not in {None, ""}:
                    # Determine if this item is a "directory" (ends with /)
                    item_key = item["Key"]
                    item_type = "directory" if item_key.endswith("/") else "file"
                    item_path = f"{self._settings.name}/{item_key}"

                    result.append({
                        "key": item_path,
                        "size": item.get("Size", 0),
                        "last_modified": item.get("LastModified"),
                        "etag": item.get("ETag", "").strip('"'),
                        "storage_class": item.get("StorageClass", ""),
                        "type": item_type,
                    })
        else:
            # Process directories (common prefixes) for non-detailed view
            for prefix in common_prefixes:
                if prefix.get("Prefix") not in {None, ""}:
                    prefix_path = f"{self._settings.name}/{prefix['Prefix']}"
                    result.append(prefix_path)

            # Process files for non-detailed view
            for item in contents:
                if item.get("Key") not in {None, ""}:
                    item_path = f"{self._settings.name}/{item['Key']}"
                    result.append(item_path)

        return result

    @staticmethod
    def find_static(detail: bool = False) -> list[str | dict[str, Any]]:
        """List objects recursively in the bucket, static method version.

        Args:
            detail (bool): If True, return detailed information including object type, else return only paths.

        Returns:
            list[Union[str, dict[str, Any]]]: List of objects in the bucket with optional detail.
        """
        return Service().find(detail=detail)

    def find(self, detail: bool = False) -> list[str | dict[str, Any]]:  # noqa: C901
        """List objects recursively in the bucket.

        Args:
            detail (bool): If True, return detailed information including object type, else return only paths.

        Returns:
            list[Union[str, dict[str, Any]]]: List of objects in the bucket with optional detail.
        """
        s3c = self._get_s3_client()
        paginator = s3c.get_paginator("list_objects_v2")
        pages = paginator.paginate(Bucket=self._settings.name)

        result: list[str | dict[str, Any]] = []
        for page in pages:
            contents = page.get("Contents", [])
            common_prefixes = page.get("CommonPrefixes", [])

            if detail:
                # Process directories (common prefixes)
                for prefix in common_prefixes:
                    if prefix.get("Prefix") not in {None, ""}:
                        prefix_path = f"{prefix['Prefix']}"
                        result.append({
                            "key": prefix_path,
                            "size": 0,
                            "last_modified": None,
                            "etag": "",
                            "storage_class": "",
                            "type": "directory",
                        })

                # Process files
                for item in contents:
                    if item.get("Key") not in {None, ""}:
                        # Determine if this item is a "directory" (ends with /)
                        item_key = item["Key"]
                        item_type = "directory" if item_key.endswith("/") else "file"
                        item_path = item_key

                        result.append({
                            "key": item_path,
                            "size": item.get("Size", 0),
                            "last_modified": item.get("LastModified"),
                            "etag": item.get("ETag", "").strip('"'),
                            "storage_class": item.get("StorageClass", ""),
                            "type": item_type,
                        })
            else:
                # Process directories (common prefixes) for non-detailed view
                for prefix in common_prefixes:
                    if prefix.get("Prefix") not in {None, ""}:
                        prefix_path = prefix["Prefix"]
                        result.append(prefix_path)

                # Process files for non-detailed view
                for item in contents:
                    if item.get("Key") not in {None, ""}:
                        item_path = item["Key"]
                        result.append(item_path)

        return result

    @staticmethod
    def delete_objects_static(keys: list[str]) -> bool:
        """Delete objects recursively in the bucket, static method version.

        Args:
            keys (list[str]): List of keys to delete.

        Returns:
            bool: True if successful, False otherwise.
        """
        return Service().delete_objects(keys)

    def delete_objects(self, keys: list[str]) -> bool:
        """Delete  objects.

        Args:
            keys (list[str]): List of keys to delete.

        Returns:
            bool: True if successful, False otherwise.
        """
        s3c = self._get_s3_client()
        for key in keys:
            logger.debug("Deleting key: %s", key)
            # Strip bucket prefix if present in key
            pruned_key = key.removeprefix(f"{self._settings.name}/")
            try:
                s3c.delete_object(Bucket=self._settings.name, Key=pruned_key)
            except ClientError as e:
                if e.response["Error"]["Code"] == "NoSuchKey":
                    logger.warning("Object with key '%s' not found", key)
                    return False
                logger.exception("Error deleting object with key '%s'", key)
                return False
        logger.info("Deleted %d objects", len(keys))
        return True



"""Settings of the system module."""

from typing import Annotated

from pydantic import Field, PlainSerializer, SecretStr
from pydantic_settings import SettingsConfigDict

from ..utils import OpaqueSettings, __env_file__, __project_name__  # noqa: TID252


class Settings(OpaqueSettings):
    """Settings."""

    model_config = SettingsConfigDict(
        env_prefix=f"{__project_name__.upper()}_SYSTEM_",
        extra="ignore",
        env_file=__env_file__,
        env_file_encoding="utf-8",
    )

    token: Annotated[
        SecretStr | None,
        PlainSerializer(func=OpaqueSettings.serialize_sensitive_info, return_type=str, when_used="always"),
        Field(
            description=(
                "Secret token to present when performing sensitive operations such as "
                "retrieving info via webservice API"
            ),
            default=None,
        ),
    ]
